import json
from dataclasses import dataclass, field
from typing import Optional, List, Iterator, Dict, Any, Mapping, Union

from pydantic import BaseModel

from agentica.model.base import Model
from agentica.model.message import Message
from agentica.model.response import ModelResponse
from agentica.tools.base import FunctionCall, get_function_call_for_tool_call
from agentica.utils.log import logger
from agentica.utils.timer import Timer

try:
    from ollama import Client as OllamaClient
    from ollama import AsyncClient as AsyncOllamaClient
except (ModuleNotFoundError, ImportError):
    raise ImportError("`ollama` not installed. Please install using `pip install ollama`")


@dataclass
class Metrics:
    input_tokens: int = 0
    output_tokens: int = 0
    total_tokens: int = 0
    time_to_first_token: Optional[float] = None
    response_timer: Timer = field(default_factory=Timer)

    def log(self):
        if self.time_to_first_token is not None:
            logger.debug(f"* Time to first token:         {self.time_to_first_token:.4f}s")
        logger.debug(f"* Time to generate response:   {self.response_timer.elapsed:.4f}s")
        logger.debug(f"* Tokens per second:           {self.output_tokens / self.response_timer.elapsed:.4f} tokens/s")
        logger.debug(f"* Input tokens:                {self.input_tokens}")
        logger.debug(f"* Output tokens:               {self.output_tokens}")
        logger.debug(f"* Total tokens:                {self.total_tokens}")


@dataclass
class MessageData:
    response_role: Optional[str] = None
    response_message: Optional[Dict[str, Any]] = None
    response_content: Any = ""
    response_content_chunk: str = ""
    tool_calls: List[Dict[str, Any]] = field(default_factory=list)
    tool_call_blocks: Any = field(default_factory=list)
    tool_call_chunk: str = ""
    in_tool_call: bool = False
    response_usage: Optional[Mapping[str, Any]] = None


class Ollama(Model):
    """
    A class for interacting with Ollama models.

    For more information, see: https://github.com/ollama/ollama/blob/main/docs/api.md
    """

    id: str = "llama3.1"
    name: str = "Ollama"
    provider: str = "Ollama"

    # Request parameters
    format: Optional[Any] = None
    options: Optional[Any] = None
    keep_alive: Optional[Union[float, str]] = None
    request_params: Optional[Dict[str, Any]] = None

    # Client parameters
    host: Optional[str] = None
    timeout: Optional[Any] = None
    client_params: Optional[Dict[str, Any]] = None

    # Ollama clients
    client: Optional[OllamaClient] = None
    async_client: Optional[AsyncOllamaClient] = None

    # Internal parameters. Not used for API requests
    # Whether to use the structured outputs with this Model.
    structured_outputs: bool = False
    # Whether the Model supports structured outputs.
    supports_structured_outputs: bool = True

    def get_client_params(self) -> Dict[str, Any]:
        client_params: Dict[str, Any] = {}
        if self.host is not None:
            client_params["host"] = self.host
        if self.timeout is not None:
            client_params["timeout"] = self.timeout
        if self.client_params is not None:
            client_params.update(self.client_params)
        return client_params

    def get_client(self) -> OllamaClient:
        """
        Returns an Ollama client.

        Returns:
            OllamaClient: An instance of the Ollama client.
        """
        if self.client is not None:
            return self.client

        return OllamaClient(**self.get_client_params())

    def get_async_client(self) -> AsyncOllamaClient:
        """
        Returns an asynchronous Ollama client.

        Returns:
            AsyncOllamaClient: An instance of the Ollama client.
        """
        if self.async_client is not None:
            return self.async_client

        return AsyncOllamaClient(**self.get_client_params())

    @property
    def request_kwargs(self) -> Dict[str, Any]:
        """
        Returns keyword arguments for API requests.

        Returns:
            Dict[str, Any]: The API kwargs for the model.
        """
        request_params: Dict[str, Any] = {}
        if self.format is not None:
            request_params["format"] = self.format
        if self.options is not None:
            request_params["options"] = self.options
        if self.keep_alive is not None:
            request_params["keep_alive"] = self.keep_alive
        if self.tools is not None:
            request_params["tools"] = self.get_tools_for_api()
            # Ensure types are valid strings
            for tool in request_params["tools"]:
                for prop, obj in tool["function"]["parameters"]["properties"].items():
                    if isinstance(obj["type"], list):
                        obj["type"] = obj["type"][0]
        if self.request_params is not None:
            request_params.update(self.request_params)
        return request_params

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the model to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the model.
        """
        model_dict = super().to_dict()
        if self.format is not None:
            model_dict["format"] = self.format
        if self.options is not None:
            model_dict["options"] = self.options
        if self.keep_alive is not None:
            model_dict["keep_alive"] = self.keep_alive
        if self.request_params is not None:
            model_dict["request_params"] = self.request_params
        return model_dict

    def format_message(self, message: Message) -> Dict[str, Any]:
        """
        Format a message into the format expected by Ollama.

        Args:
            message (Message): The message to format.

        Returns:
            Dict[str, Any]: The formatted message.
        """
        _message: Dict[str, Any] = {
            "role": message.role,
            "content": message.content,
        }
        if message.role == "user":
            if message.images is not None:
                _message["images"] = message.images
        return _message

    def invoke(self, messages: List[Message]) -> Mapping[str, Any]:
        """
        Send a chat request to the Ollama API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            Mapping[str, Any]: The response from the API.
        """
        request_kwargs = self.request_kwargs
        if self.response_format is not None and self.structured_outputs:
            if isinstance(self.response_format, type) and issubclass(self.response_format, BaseModel):
                logger.debug("Using structured outputs")
                format_schema = self.response_format.model_json_schema()
                if "format" not in request_kwargs:
                    request_kwargs["format"] = format_schema

        return self.get_client().chat(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            **request_kwargs,
        )  # type: ignore

    async def ainvoke(self, messages: List[Message]) -> Mapping[str, Any]:
        """
        Sends an asynchronous chat request to the Ollama API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            Mapping[str, Any]: The response from the API.
        """
        request_kwargs = self.request_kwargs
        if self.response_format is not None and self.structured_outputs:
            if isinstance(self.response_format, type) and issubclass(self.response_format, BaseModel):
                logger.debug("Using structured outputs")
                format_schema = self.response_format.model_json_schema()
                if "format" not in request_kwargs:
                    request_kwargs["format"] = format_schema

        return await self.get_async_client().chat(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            **request_kwargs,
        )  # type: ignore

    def invoke_stream(self, messages: List[Message]) -> Iterator[Mapping[str, Any]]:
        """
        Sends a streaming chat request to the Ollama API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            Iterator[Mapping[str, Any]]: An iterator of chunks from the API.
        """
        yield from self.get_client().chat(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            stream=True,
            **self.request_kwargs,
        )  # type: ignore

    async def ainvoke_stream(self, messages: List[Message]) -> Any:
        """
        Sends an asynchronous streaming chat completion request to the Ollama API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            Any: An asynchronous iterator of chunks from the API.
        """
        async_stream = await self.get_async_client().chat(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            stream=True,
            **self.request_kwargs,
        )
        async for chunk in async_stream:  # type: ignore
            yield chunk

    def handle_tool_calls(
        self,
        assistant_message: Message,
        messages: List[Message],
        model_response: ModelResponse,
    ) -> Optional[ModelResponse]:
        """
        Handle tool calls in the assistant message.

        Args:
            assistant_message (Message): The assistant message.
            messages (List[Message]): The list of messages.
            model_response (ModelResponse): The model response.

        Returns:
            Optional[ModelResponse]: The model response.
        """
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            model_response.content = assistant_message.get_content_string()
            model_response.content += "\n\n"
            function_calls_to_run = self.get_function_calls_to_run(assistant_message, messages)
            function_call_results: List[Message] = []

            if self.show_tool_calls:
                if len(function_calls_to_run) == 1:
                    model_response.content += f" - Running: {function_calls_to_run[0].get_call_str()}\n\n"
                elif len(function_calls_to_run) > 1:
                    model_response.content += "Running:"
                    for _f in function_calls_to_run:
                        model_response.content += f"\n - {_f.get_call_str()}"
                    model_response.content += "\n\n"

            for _ in self.run_function_calls(
                function_calls=function_calls_to_run,
                function_call_results=function_call_results,
            ):
                pass

            self.format_function_call_results(function_call_results, messages)

            return model_response
        return None

    def update_usage_metrics(
        self,
        assistant_message: Message,
        metrics: Metrics,
        response: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """
        Update usage metrics for the assistant message.

        Args:
            assistant_message (Message): The assistant message.
            metrics (Optional[Metrics]): The metrics for this response.
            response (Optional[Mapping[str, Any]]): The response from Ollama.
        """
        # Update time taken to generate response
        assistant_message.metrics["time"] = metrics.response_timer.elapsed
        self.metrics.setdefault("response_times", []).append(metrics.response_timer.elapsed)
        if response:
            metrics.input_tokens = response.get("prompt_eval_count", 0)
            metrics.output_tokens = response.get("eval_count", 0)
            metrics.total_tokens = metrics.input_tokens + metrics.output_tokens

            if metrics.input_tokens is not None:
                assistant_message.metrics["input_tokens"] = metrics.input_tokens
                self.metrics["input_tokens"] = self.metrics.get("input_tokens", 0) + metrics.input_tokens
            if metrics.output_tokens is not None:
                assistant_message.metrics["output_tokens"] = metrics.output_tokens
                self.metrics["output_tokens"] = self.metrics.get("output_tokens", 0) + metrics.output_tokens
            if metrics.total_tokens is not None:
                assistant_message.metrics["total_tokens"] = metrics.total_tokens
                self.metrics["total_tokens"] = self.metrics.get("total_tokens", 0) + metrics.total_tokens
            if metrics.time_to_first_token is not None:
                assistant_message.metrics["time_to_first_token"] = metrics.time_to_first_token
                self.metrics.setdefault("time_to_first_token", []).append(metrics.time_to_first_token)

    def get_function_calls_to_run(self, assistant_message: Message, messages: List[Message]) -> List[FunctionCall]:
        """
        Get the function calls to run from the assistant message.

        Args:
            assistant_message (Message): The assistant message.
            messages (List[Message]): The list of messages.

        Returns:
            List[FunctionCall]: The list of function calls to run.
        """
        function_calls_to_run: List[FunctionCall] = []
        if assistant_message.tool_calls is not None:
            for tool_call in assistant_message.tool_calls:
                _function_call = get_function_call_for_tool_call(tool_call, self.functions)
                if _function_call is None:
                    messages.append(Message(role="user", content="Could not find function to call."))
                    continue
                if _function_call.error is not None:
                    messages.append(Message(role="user", content=_function_call.error))
                    continue
                function_calls_to_run.append(_function_call)
        return function_calls_to_run

    def format_function_call_results(self, function_call_results: List[Message], messages: List[Message]) -> None:
        """
        Format the function call results and append them to the messages.

        Args:
            function_call_results (List[Message]): The list of function call results.
            messages (List[Message]): The list of messages.
        """
        if len(function_call_results) > 0:
            for _fcr in function_call_results:
                messages.append(_fcr)

    def create_assistant_message(self, response: Mapping[str, Any], metrics: Metrics) -> Message:
        """
        Create an assistant message from the response.

        Args:
            response: The response from Ollama.
            metrics: The metrics for this response.

        Returns:
            Message: The assistant message.
        """
        message_data = MessageData()

        message_data.response_message = response.get("message")
        if message_data.response_message:
            message_data.response_content = message_data.response_message.get("content")
            message_data.response_role = message_data.response_message.get("role")
            message_data.tool_call_blocks = message_data.response_message.get("tool_calls")

        assistant_message = Message(
            role=message_data.response_role or "assistant",
            content=message_data.response_content,
        )
        if message_data.tool_call_blocks is not None:
            for block in message_data.tool_call_blocks:
                tool_call = block.get("function")
                tool_name = tool_call.get("name")
                tool_args = tool_call.get("arguments")

                function_def = {
                    "name": tool_name,
                    "arguments": json.dumps(tool_args) if tool_args is not None else None,
                }
                message_data.tool_calls.append({"type": "function", "function": function_def})

        if message_data.tool_calls is not None:
            assistant_message.tool_calls = message_data.tool_calls

        # Update metrics
        self.update_usage_metrics(assistant_message=assistant_message, metrics=metrics, response=response)
        return assistant_message

    def response(self, messages: List[Message]) -> ModelResponse:
        """
        Generate a response from Ollama.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            ModelResponse: The model response.
        """
        self._log_messages(messages)
        model_response = ModelResponse()
        metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        response: Mapping[str, Any] = self.invoke(messages=messages)
        metrics.response_timer.stop()

        # -*- Parse structured outputs
        try:
            if (
                self.response_format is not None
                and self.structured_outputs
                and issubclass(self.response_format, BaseModel)
            ):
                parsed_object = self.response_format.model_validate_json(response.get("message", {}).get("content", ""))
                if parsed_object is not None:
                    model_response.parsed = parsed_object.model_dump_json()
        except Exception as e:
            logger.warning(f"Error parsing structured outputs: {e}")

        # -*- Create assistant message
        assistant_message = self.create_assistant_message(response=response, metrics=metrics)

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Update model response with assistant message content and audio
        if assistant_message.content is not None:
            # add the content to the model response
            model_response.content = assistant_message.get_content_string()
        if assistant_message.audio is not None:
            # add the audio to the model response
            model_response.audio = assistant_message.audio

        # -*- Handle tool calls
        if (
            self.handle_tool_calls(
                assistant_message=assistant_message, messages=messages, model_response=model_response
            )
            is not None
        ):
            return self.handle_post_tool_call_messages(messages=messages, model_response=model_response)
        return model_response

    async def aresponse(self, messages: List[Message]) -> ModelResponse:
        """
        Generate an asynchronous response from Ollama.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            ModelResponse: The model response.
        """
        self._log_messages(messages)
        model_response = ModelResponse()
        metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        response: Mapping[str, Any] = await self.ainvoke(messages=messages)
        metrics.response_timer.stop()

        # -*- Parse structured outputs
        try:
            if (
                self.response_format is not None
                and self.structured_outputs
                and issubclass(self.response_format, BaseModel)
            ):
                parsed_object = self.response_format.model_validate_json(response.get("message", {}).get("content", ""))
                if parsed_object is not None:
                    model_response.parsed = parsed_object.model_dump_json()
        except Exception as e:
            logger.warning(f"Error parsing structured outputs: {e}")

        # -*- Create assistant message
        assistant_message = self.create_assistant_message(response=response, metrics=metrics)

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Update model response with assistant message content and audio
        if assistant_message.content is not None:
            # add the content to the model response
            model_response.content = assistant_message.get_content_string()
        if assistant_message.audio is not None:
            # add the audio to the model response
            model_response.audio = assistant_message.audio

        # -*- Handle tool calls
        if (
            self.handle_tool_calls(
                assistant_message=assistant_message, messages=messages, model_response=model_response
            )
            is not None
        ):
            return await self.ahandle_post_tool_call_messages(messages=messages, model_response=model_response)

        return model_response

    def handle_stream_tool_calls(
        self,
        assistant_message: Message,
        messages: List[Message],
    ) -> Iterator[ModelResponse]:
        """
        Handle tool calls for response stream.

        Args:
            assistant_message (Message): The assistant message.
            messages (List[Message]): The list of messages.

        Returns:
            Iterator[ModelResponse]: An iterator of the model response.
        """
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            yield ModelResponse(content="\n\n")
            function_calls_to_run = self.get_function_calls_to_run(assistant_message, messages)
            function_call_results: List[Message] = []

            if self.show_tool_calls:
                if len(function_calls_to_run) == 1:
                    yield ModelResponse(content=f" - Running: {function_calls_to_run[0].get_call_str()}\n\n")
                elif len(function_calls_to_run) > 1:
                    yield ModelResponse(content="Running:")
                    for _f in function_calls_to_run:
                        yield ModelResponse(content=f"\n - {_f.get_call_str()}")
                    yield ModelResponse(content="\n\n")

            for intermediate_model_response in self.run_function_calls(
                function_calls=function_calls_to_run, function_call_results=function_call_results
            ):
                yield intermediate_model_response

            self.format_function_call_results(function_call_results, messages)

    def response_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        """
        Generate a streaming response from Ollama.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            Iterator[ModelResponse]: An iterator of the model responses.
        """
        self._log_messages(messages)
        message_data = MessageData()
        metrics: Metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        for response in self.invoke_stream(messages=messages):
            # logger.debug(f"Response: {response}")
            message_data.response_message = response.get("message", {})
            if message_data.response_message:
                metrics.output_tokens += 1
                if metrics.output_tokens == 1:
                    metrics.time_to_first_token = metrics.response_timer.elapsed

                message_data.response_content_chunk = message_data.response_message.get("content", "")
                if message_data.response_content_chunk is not None and message_data.response_content_chunk != "":
                    message_data.response_content += message_data.response_content_chunk
                    yield ModelResponse(content=message_data.response_content_chunk)

                message_data.tool_call_blocks = message_data.response_message.get("tool_calls")  # type: ignore
                if message_data.tool_call_blocks is not None:
                    for block in message_data.tool_call_blocks:
                        tool_call = block.get("function")
                        tool_name = tool_call.get("name")
                        tool_args = tool_call.get("arguments")
                        function_def = {
                            "name": tool_name,
                            "arguments": json.dumps(tool_args) if tool_args is not None else None,
                        }
                        message_data.tool_calls.append({"type": "function", "function": function_def})

            if response.get("done"):
                message_data.response_usage = response
        metrics.response_timer.stop()

        # -*- Create assistant message
        assistant_message = Message(role="assistant", content=message_data.response_content)

        if len(message_data.tool_calls) > 0:
            assistant_message.tool_calls = message_data.tool_calls

        # -*- Update usage metrics
        self.update_usage_metrics(
            assistant_message=assistant_message, metrics=metrics, response=message_data.response_usage
        )

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Handle tool calls
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            yield from self.handle_stream_tool_calls(assistant_message, messages)
            yield from self.handle_post_tool_call_messages_stream(messages=messages)

    async def aresponse_stream(self, messages: List[Message]) -> Any:
        """
        Generate an asynchronous streaming response from Ollama.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            Any: An asynchronous iterator of the model responses.
        """
        self._log_messages(messages)
        message_data = MessageData()
        metrics: Metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        async for response in self.ainvoke_stream(messages=messages):
            message_data.response_message = response.get("message", {})
            if message_data.response_message:
                metrics.output_tokens += 1
                if metrics.output_tokens == 1:
                    metrics.time_to_first_token = metrics.response_timer.elapsed

                message_data.response_content_chunk = message_data.response_message.get("content", "")
                if message_data.response_content_chunk is not None and message_data.response_content_chunk != "":
                    message_data.response_content += message_data.response_content_chunk
                    yield ModelResponse(content=message_data.response_content_chunk)

                message_data.tool_call_blocks = message_data.response_message.get("tool_calls")
                if message_data.tool_call_blocks is not None:
                    for block in message_data.tool_call_blocks:
                        tool_call = block.get("function")
                        tool_name = tool_call.get("name")
                        tool_args = tool_call.get("arguments")
                        function_def = {
                            "name": tool_name,
                            "arguments": json.dumps(tool_args) if tool_args is not None else None,
                        }
                        message_data.tool_calls.append({"type": "function", "function": function_def})

            if response.get("done"):
                message_data.response_usage = response
        metrics.response_timer.stop()

        # -*- Create assistant message
        assistant_message = Message(role="assistant", content=message_data.response_content)

        if len(message_data.tool_calls) > 0:
            assistant_message.tool_calls = message_data.tool_calls

        # -*- Update usage metrics
        self.update_usage_metrics(
            assistant_message=assistant_message, metrics=metrics, response=message_data.response_usage
        )

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Handle tool calls
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            for tool_call_response in self.handle_stream_tool_calls(assistant_message, messages):
                yield tool_call_response
            async for post_tool_call_response in self.ahandle_post_tool_call_messages_stream(messages=messages):
                yield post_tool_call_response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by Charles on 2018/10/10
# Function:

import requests
from bs4 import BeautifulSoup
import json
from typing import Optional, List, Dict, Any

from agentica.tools.base import Tool
from agentica.utils.log import logger

ABSTRACT_MAX_LENGTH = 300  # abstract max length

user_agents = [
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.106 Safari/537.36',
    'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko)'
    ' Ubuntu Chromium/49.0.2623.108 Chrome/49.0.2623.108 Safari/537.36',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; pt-BR) AppleWebKit/533.3 '
    '(KHTML, like Gecko)  QtWeb Internet Browser/3.7 http://www.QtWeb.net',
    'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/41.0.2228.0 Safari/537.36',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/532.2 (KHTML, '
    'like Gecko) ChromePlus/4.0.222.3 Chrome/4.0.222.3 Safari/532.2',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.4pre) '
    'Gecko/20070404 K-Ninja/2.1.3',
    'Mozilla/5.0 (Future Star Technologies Corp.; Star-Blade OS; x86_64; U; '
    'en-US) iNet Browser 4.7',
    'Mozilla/5.0 (Windows; U; Windows NT 6.1; rv:2.2) Gecko/20110201',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.13) '
    'Gecko/20080414 Firefox/2.0.0.13 Pogo/2.0.0.13.6866'
]

# 请求头信息
HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
    "Content-Type": "application/x-www-form-urlencoded",
    "User-Agent": 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.106 Safari/537.36',
    "Referer": "https://www.baidu.com/",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "zh-CN,zh;q=0.9"
}

baidu_host_url = "https://www.baidu.com"
baidu_search_url = "https://www.baidu.com/s?ie=utf-8&tn=baidu&wd="

session = requests.Session()
session.headers = HEADERS


def search(keyword, num_results=10, debug=False):
    """
    通过关键字进行搜索
    :param keyword: 关键字
    :param num_results： 指定返回的结果个数
    :param debug: 是否开启debug模式
    :return: 结果列表
    """
    if not keyword:
        return None

    list_result = []
    page = 1

    # 起始搜索的url
    next_url = baidu_search_url + keyword

    # 循环遍历每一页的搜索结果，并返回下一页的url
    while len(list_result) < num_results:
        data, next_url = parse_html(next_url, rank_start=len(list_result))
        if data:
            list_result += data
            if debug:
                logger.debug("---searching[{}], finish parsing page {}, results number={}: ".format(
                    keyword, page, len(data)))
                for d in data:
                    logger.debug(str(d))

        if not next_url:
            if debug:
                logger.debug(u"already search the last page。")
            break
        page += 1

    if debug:
        logger.debug("\n---search [{}] finished. total results number={}！".format(keyword, len(list_result)))
    return list_result[: num_results] if len(list_result) > num_results else list_result


def parse_html(url, rank_start=0, debug=0):
    """
    解析处理结果
    :param url: 需要抓取的 url
    :return:  结果列表，下一页的url
    """
    try:
        res = session.get(url=url)
        res.encoding = "utf-8"
        root = BeautifulSoup(res.text, "lxml")

        list_data = []
        div_contents = root.find("div", id="content_left")
        for div in div_contents.contents:
            if type(div) != type(div_contents):
                continue

            class_list = div.get("class", [])
            if not class_list:
                continue

            if "c-container" not in class_list:
                continue

            title = ''
            url = ''
            abstract = ''
            try:
                # 遍历所有找到的结果，取得标题和概要内容（50字以内）
                if "xpath-log" in class_list:
                    if div.h3:
                        title = div.h3.text.strip()
                        url = div.h3.a['href'].strip()
                    else:
                        title = div.text.strip().split("\n", 1)[0]
                        if div.a:
                            url = div.a['href'].strip()

                    if div.find("div", class_="c-abstract"):
                        abstract = div.find("div", class_="c-abstract").text.strip()
                    elif div.div:
                        abstract = div.div.text.strip()
                    else:
                        abstract = div.text.strip().split("\n", 1)[1].strip()
                elif "result-op" in class_list:
                    if div.h3:
                        title = div.h3.text.strip()
                        url = div.h3.a['href'].strip()
                    else:
                        title = div.text.strip().split("\n", 1)[0]
                        url = div.a['href'].strip()
                    if div.find("div", class_="c-abstract"):
                        abstract = div.find("div", class_="c-abstract").text.strip()
                    elif div.div:
                        abstract = div.div.text.strip()
                    else:
                        # abstract = div.text.strip()
                        abstract = div.text.strip().split("\n", 1)[1].strip()
                else:
                    if div.get("tpl", "") != "se_com_default":
                        if div.get("tpl", "") == "se_st_com_abstract":
                            if len(div.contents) >= 1:
                                title = div.h3.text.strip()
                                if div.find("div", class_="c-abstract"):
                                    abstract = div.find("div", class_="c-abstract").text.strip()
                                elif div.div:
                                    abstract = div.div.text.strip()
                                else:
                                    abstract = div.text.strip()
                        else:
                            if len(div.contents) >= 2:
                                if div.h3:
                                    title = div.h3.text.strip()
                                    url = div.h3.a['href'].strip()
                                else:
                                    title = div.contents[0].text.strip()
                                    url = div.h3.a['href'].strip()
                                # abstract = div.contents[-1].text
                                if div.find("div", class_="c-abstract"):
                                    abstract = div.find("div", class_="c-abstract").text.strip()
                                elif div.div:
                                    abstract = div.div.text.strip()
                                else:
                                    abstract = div.text.strip()
                    else:
                        if div.h3:
                            title = div.h3.text.strip()
                            url = div.h3.a['href'].strip()
                        else:
                            title = div.contents[0].text.strip()
                            url = div.h3.a['href'].strip()
                        if div.find("div", class_="c-abstract"):
                            abstract = div.find("div", class_="c-abstract").text.strip()
                        elif div.div:
                            abstract = div.div.text.strip()
                        else:
                            abstract = div.text.strip()
            except Exception as e:
                if debug:
                    logger.debug("catch exception duration parsing page html, e={}".format(e))
                continue

            if ABSTRACT_MAX_LENGTH and len(abstract) > ABSTRACT_MAX_LENGTH:
                abstract = abstract[:ABSTRACT_MAX_LENGTH]

            rank_start += 1
            list_data.append({"title": title, "abstract": abstract, "url": url, "rank": rank_start})

        # 找到下一页按钮
        next_btn = root.find_all("a", class_="n")

        # 已经是最后一页了，没有下一页了，此时只返回数据不再获取下一页的链接
        if len(next_btn) <= 0 or u"上一页" in next_btn[-1].text:
            return list_data, None

        next_url = baidu_host_url + next_btn[-1]["href"]
        return list_data, next_url
    except Exception as e:
        if debug:
            logger.debug(u"catch exception duration parsing page html, e：{}".format(e))
        return None, None


class BaiduSearchTool(Tool):
    """
    BaiduSearch is a toolkit for searching Baidu easily.

    Args:
        fixed_max_results (Optional[int]): A fixed number of maximum results.
        fixed_language (Optional[str]): A fixed language for the search results.
        headers (Optional[Any]): Headers to be used in the search request.
        debug (Optional[bool]): Enable debug output.
    """

    def __init__(
            self,
            fixed_max_results: Optional[int] = 5,
            fixed_language: Optional[str] = "zh",
            headers: Optional[Any] = None,
            timeout: Optional[int] = 10,
            debug: Optional[bool] = False,
    ):
        super().__init__(name="baidusearch")
        self.fixed_max_results = fixed_max_results
        self.fixed_language = fixed_language
        self.headers = headers
        self.timeout = timeout
        self.debug = debug
        self.register(self.baidu_search)

    def baidu_search(self, query: str, max_results: int = 5, language: str = "zh") -> str:
        """Execute Baidu search and return results

        Args:
            query (str): Search keyword
            max_results (int, optional): Maximum number of results to return, default 5
            language (str, optional): Search language, default is 'zh'

        Returns:
            str: A JSON formatted string containing the search results.
        """
        max_results = self.fixed_max_results or max_results
        language = self.fixed_language or language

        if len(language) != 2:
            try:
                from pycountry import pycountry
                language = pycountry.languages.lookup(language).alpha_2
            except ImportError:
                raise ImportError("`pycountry` not installed. Please install using `pip install pycountry`")
            except LookupError:
                language = "zh"

        logger.info(f"Searching Baidu [{language}] for: {query}")
        results = search(keyword=query, num_results=max_results, debug=self.debug)

        res: List[Dict[str, str]] = []
        for idx, item in enumerate(results, 1):
            res.append(
                {
                    "title": item.get("title", ""),
                    "url": item.get("url", ""),
                    "abstract": item.get("abstract", ""),
                    "rank": str(idx),
                }
            )
        return json.dumps(res, indent=2, ensure_ascii=False)


if __name__ == '__main__':
    keyword = "NBA"
    results = search(keyword, num_results=5)
    print(results)
    s = BaiduSearchTool()
    r = s.baidu_search("北京的新闻top3")
    print(r)



# -*- coding: utf-8 -*-
"""
@author: XuMing(xuming624@qq.com)
@description: Tools for code analysis, formatting, linting, refactoring, and more.
"""
import ast
import json
import os
import re
import subprocess
from pathlib import Path
from typing import Optional, Dict, Any, List, Tuple, Union

from agentica.tools.base import Tool
from agentica.utils.log import logger


class CodeTool(Tool):
    """
    A toolkit for code operations, including analysis, formatting, and linting.
    Essential for building code editor functionality.
    """

    def __init__(
            self,
            work_dir: Optional[str] = None,
            analyze_code: bool = True,
            format_code: bool = True,
            run_code: bool = True,
            lint_code: bool = True,
            find_symbols: bool = True,
            get_code_outline: bool = True,
    ):
        """
        Initialize the CodeTool.

        Args:
            work_dir: The working directory for code operations. Defaults to current directory.
            analyze_code: Whether to include the analyze_code function.
            format_code: Whether to include the format_code function.
            run_code: Whether to include the run_code function.
            lint_code: Whether to include the lint_code function.
            find_symbols: Whether to include the find_symbols function.
            get_code_outline: Whether to include the get_code_outline function.
        """
        super().__init__(name="code_tool")
        self.work_dir = Path(work_dir) if work_dir else Path.cwd()
        
        if analyze_code:
            self.register(self.analyze_code)
        if format_code:
            self.register(self.format_code)
        if run_code:
            self.register(self.run_code)
        if lint_code:
            self.register(self.lint_code)
        if find_symbols:
            self.register(self.find_symbols)
        if get_code_outline:
            self.register(self.get_code_outline)

    def _resolve_path(self, file_path: str) -> Path:
        """Resolves a file path, making it absolute if it's relative."""
        path = Path(file_path)
        if not path.is_absolute():
            path = self.work_dir.joinpath(path)
        return path

    def _run_command(self, cmd: List[str], cwd: str = None, timeout: int = 60) -> Tuple[str, str, int]:
        """Run a shell command and return stdout, stderr, and return code."""
        try:
            process = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                cwd=cwd or str(self.work_dir)
            )
            stdout, stderr = process.communicate(timeout=timeout)
            return stdout, stderr, process.returncode
        except subprocess.TimeoutExpired:
            process.kill()
            return "", "Command timed out", 1
        except Exception as e:
            return "", f"Error executing command: {str(e)}", 1

    def analyze_code(self, file_path: str) -> str:
        """Analyze Python code and return useful information about it.

        Args:
            file_path (str): Path to the Python file to analyze.

        Example:
            from agentica.tools.code_tool import CodeTool
            analyzer = CodeTool()
            result = analyzer.analyze_code("example.py")
            print(result)

        Returns:
            str: JSON-formatted analysis of the code including functions, classes, imports, etc.
        """
        try:
            path = self._resolve_path(file_path)
            
            if not path.exists():
                return f"Error: File not found: {file_path}"
            
            if not file_path.endswith('.py'):
                return f"Error: Only Python files are supported for analysis. File: {file_path}"
            
            with open(path, 'r', encoding='utf-8') as f:
                code = f.read()
            
            try:
                tree = ast.parse(code)
            except SyntaxError as e:
                return f"Error: Syntax error in {file_path}: {str(e)}"
            
            analysis = {
                'imports': [],
                'functions': [],
                'classes': [],
                'global_variables': [],
                'total_lines': len(code.splitlines()),
                'complexity_estimate': 'low'  # Default value
            }
            
            # Analyze imports
            for node in ast.walk(tree):
                if isinstance(node, ast.Import):
                    for name in node.names:
                        analysis['imports'].append({
                            'name': name.name,
                            'alias': name.asname
                        })
                elif isinstance(node, ast.ImportFrom):
                    module = node.module or ''
                    for name in node.names:
                        analysis['imports'].append({
                            'name': f"{module}.{name.name}" if module else name.name,
                            'alias': name.asname,
                            'from_import': True
                        })
            
            # Analyze functions
            for node in ast.walk(tree):
                if isinstance(node, ast.FunctionDef):
                    args = []
                    for arg in node.args.args:
                        args.append(arg.arg)
                    
                    # Extract docstring if it exists
                    docstring = ast.get_docstring(node) or ''
                    
                    analysis['functions'].append({
                        'name': node.name,
                        'args': args,
                        'line_number': node.lineno,
                        'docstring': docstring,
                        'decorator_list': [d.id if isinstance(d, ast.Name) else "complex_decorator" for d in node.decorator_list]
                    })
            
            # Analyze classes
            for node in ast.walk(tree):
                if isinstance(node, ast.ClassDef):
                    methods = []
                    for child in node.body:
                        if isinstance(child, ast.FunctionDef):
                            methods.append(child.name)
                    
                    # Get base classes
                    bases = []
                    for base in node.bases:
                        if isinstance(base, ast.Name):
                            bases.append(base.id)
                        elif isinstance(base, ast.Attribute):
                            bases.append(f"{base.value.id}.{base.attr}" if hasattr(base.value, 'id') else "complex_base")
                    
                    # Extract docstring if it exists
                    docstring = ast.get_docstring(node) or ''
                    
                    analysis['classes'].append({
                        'name': node.name,
                        'methods': methods,
                        'bases': bases,
                        'line_number': node.lineno,
                        'docstring': docstring
                    })
            
            # Analyze global variables
            for node in ast.walk(tree):
                if isinstance(node, ast.Assign) and all(isinstance(target, ast.Name) for target in node.targets):
                    if isinstance(node.value, ast.Constant):
                        value = node.value.value if hasattr(node.value, 'value') else "complex_value"
                    else:
                        value = "complex_value"
                    
                    for target in node.targets:
                        if isinstance(target, ast.Name):
                            analysis['global_variables'].append({
                                'name': target.id,
                                'value': str(value),
                                'line_number': node.lineno
                            })
            
            # Estimate complexity
            complexity = 'low'
            if len(analysis['functions']) + len(analysis['classes']) > 10:
                complexity = 'medium'
            if len(analysis['functions']) + len(analysis['classes']) > 20:
                complexity = 'high'
            if analysis['total_lines'] > 500:
                complexity = 'high'
            analysis['complexity_estimate'] = complexity
            
            return json.dumps(analysis, indent=2)
            
        except Exception as e:
            error_msg = f"Error analyzing code in {file_path}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def format_code(self, file_path: str, formatter: str = "black") -> str:
        """Format a code file using the specified formatter.

        Args:
            file_path (str): Path to the file to format.
            formatter (str): Formatter to use (black, autopep8, yapf, prettier). Defaults to black.

        Example:
            from agentica.tools.code_tool import CodeTool
            formatter = CodeTool()
            result = formatter.format_code("example.py", formatter="black")
            print(result)

        Returns:
            str: Result of the formatting operation.
        """
        try:
            path = self._resolve_path(file_path)
            
            if not path.exists():
                return f"Error: File not found: {file_path}"
            
            # Determine file type and appropriate formatter
            file_ext = path.suffix.lower()
            
            # Override formatter based on file extension if needed
            if formatter == "black" and file_ext != ".py":
                if file_ext in [".js", ".jsx", ".ts", ".tsx", ".json", ".html", ".css"]:
                    formatter = "prettier"
            
            cmd = []
            if formatter == "black":
                cmd = ["black", str(path)]
            elif formatter == "autopep8":
                cmd = ["autopep8", "--in-place", str(path)]
            elif formatter == "yapf":
                cmd = ["yapf", "--in-place", str(path)]
            elif formatter == "prettier":
                cmd = ["prettier", "--write", str(path)]
            else:
                return f"Error: Unsupported formatter: {formatter}"
            
            # Check if formatter is installed
            try:
                subprocess.run([cmd[0], "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False)
            except FileNotFoundError:
                return f"Error: {formatter} is not installed. Please install it first."
            
            # Run formatter
            stdout, stderr, returncode = self._run_command(cmd)
            
            if returncode == 0:
                return f"Successfully formatted {file_path} with {formatter}"
            else:
                return f"Error formatting {file_path} with {formatter}: {stderr}"
            
        except Exception as e:
            error_msg = f"Error formatting {file_path}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def run_code(self, file_path: str, args: str = "", timeout: int = 30) -> str:
        """Run a code file and return the output.

        Args:
            file_path (str): Path to the file to run.
            args (str): Command-line arguments to pass to the program.
            timeout (int): Maximum execution time in seconds.

        Example:
            from agentica.tools.code_tool import CodeTool
            runner = CodeTool()
            result = runner.run_code("example.py", args="--verbose")
            print(result)

        Returns:
            str: Output from running the code.
        """
        try:
            path = self._resolve_path(file_path)
            
            if not path.exists():
                return f"Error: File not found: {file_path}"
            
            # Determine how to run the file based on extension
            file_ext = path.suffix.lower()
            
            cmd = []
            if file_ext == ".py":
                cmd = ["python", str(path)]
            elif file_ext in [".js", ".mjs"]:
                cmd = ["node", str(path)]
            elif file_ext == ".sh":
                cmd = ["bash", str(path)]
            elif file_ext == ".rb":
                cmd = ["ruby", str(path)]
            elif file_ext == ".pl":
                cmd = ["perl", str(path)]
            elif file_ext in [".go"]:
                cmd = ["go", "run", str(path)]
            elif path.is_file() and os.access(path, os.X_OK):
                # File is executable
                cmd = [str(path)]
            else:
                return f"Error: Unsupported file type or file is not executable: {file_path}"
            
            # Add arguments if provided
            if args:
                cmd.extend(args.split())
            
            # Run the command
            stdout, stderr, returncode = self._run_command(cmd, timeout=timeout)
            
            result = f"Exit code: {returncode}\n\nStandard Output:\n{stdout}\n\nStandard Error:\n{stderr}"
            return result
            
        except Exception as e:
            error_msg = f"Error running {file_path}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def lint_code(self, file_path: str, linter: str = "auto") -> str:
        """Lint a code file and return issues found.

        Args:
            file_path (str): Path to the file to lint.
            linter (str): Linter to use (auto, pylint, flake8, eslint). Defaults to auto.

        Example:
            from agentica.tools.code_tool import CodeTool
            linter = CodeTool()
            result = linter.lint_code("example.py")
            print(result)

        Returns:
            str: Result of the linting operation with any issues found.
        """
        try:
            path = self._resolve_path(file_path)
            
            if not path.exists():
                return f"Error: File not found: {file_path}"
            
            # Determine file type and appropriate linter
            file_ext = path.suffix.lower()
            
            if linter == "auto":
                if file_ext == ".py":
                    linter = "pylint"
                elif file_ext in [".js", ".jsx", ".ts", ".tsx"]:
                    linter = "eslint"
                else:
                    return f"Error: Could not determine appropriate linter for {file_path}. Please specify a linter."
            
            cmd = []
            if linter == "pylint":
                cmd = ["pylint", str(path)]
            elif linter == "flake8":
                cmd = ["flake8", str(path)]
            elif linter == "eslint":
                cmd = ["eslint", str(path)]
            else:
                return f"Error: Unsupported linter: {linter}"
            
            # Check if linter is installed
            try:
                subprocess.run([cmd[0], "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False)
            except FileNotFoundError:
                return f"Error: {linter} is not installed. Please install it first."
            
            # Run linter
            stdout, stderr, returncode = self._run_command(cmd)
            
            if stdout or stderr:
                return f"Linting results for {file_path} with {linter}:\n\n{stdout}\n{stderr}"
            elif returncode == 0:
                return f"No issues found in {file_path} with {linter}"
            else:
                return f"Error linting {file_path} with {linter}: Return code {returncode}"
            
        except Exception as e:
            error_msg = f"Error linting {file_path}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def find_symbols(self, file_path: str, symbol_type: str = "all", pattern: str = "") -> str:
        """Find and list symbols (functions, classes, variables) in a code file.

        Args:
            file_path (str): Path to the file to analyze.
            symbol_type (str): Type of symbols to find (all, function, class, variable).
            pattern (str): Regular expression pattern to filter symbols by name.

        Example:
            from agentica.tools.code_tool import CodeTool
            finder = CodeTool()
            result = finder.find_symbols("example.py", symbol_type="function", pattern="test_")
            print(result)

        Returns:
            str: JSON-formatted list of found symbols.
        """
        try:
            path = self._resolve_path(file_path)
            
            if not path.exists():
                return f"Error: File not found: {file_path}"
            
            if not file_path.endswith('.py'):
                return f"Error: Only Python files are supported for symbol finding. File: {file_path}"
            
            with open(path, 'r', encoding='utf-8') as f:
                code = f.read()
            
            try:
                tree = ast.parse(code)
            except SyntaxError as e:
                return f"Error: Syntax error in {file_path}: {str(e)}"
            
            symbols = []
            
            # Helper to check pattern
            def matches_pattern(name):
                if not pattern:
                    return True
                try:
                    return bool(re.search(pattern, name))
                except re.error:
                    return pattern in name
            
            # Find functions
            if symbol_type in ["all", "function"]:
                for node in ast.walk(tree):
                    if isinstance(node, ast.FunctionDef) and matches_pattern(node.name):
                        args = [arg.arg for arg in node.args.args]
                        symbols.append({
                            'type': 'function',
                            'name': node.name,
                            'line': node.lineno,
                            'args': args,
                            'docstring': ast.get_docstring(node) or ''
                        })
            
            # Find classes
            if symbol_type in ["all", "class"]:
                for node in ast.walk(tree):
                    if isinstance(node, ast.ClassDef) and matches_pattern(node.name):
                        symbols.append({
                            'type': 'class',
                            'name': node.name,
                            'line': node.lineno,
                            'docstring': ast.get_docstring(node) or '',
                            'methods': [n.name for n in node.body if isinstance(n, ast.FunctionDef)]
                        })
            
            # Find variables
            if symbol_type in ["all", "variable"]:
                for node in ast.walk(tree):
                    if isinstance(node, ast.Assign):
                        for target in node.targets:
                            if isinstance(target, ast.Name) and matches_pattern(target.id):
                                symbols.append({
                                    'type': 'variable',
                                    'name': target.id,
                                    'line': node.lineno
                                })
            
            # Sort by line number
            symbols.sort(key=lambda x: x['line'])
            
            return json.dumps(symbols, indent=2)
            
        except Exception as e:
            error_msg = f"Error finding symbols in {file_path}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def get_code_outline(self, file_path: str) -> str:
        """Generate a hierarchical outline of a code file's structure.

        Args:
            file_path (str): Path to the file to outline.

        Example:
            from agentica.tools.code_tool import CodeTool
            outliner = CodeTool()
            result = outliner.get_code_outline("example.py")
            print(result)

        Returns:
            str: Markdown-formatted outline of the code structure.
        """
        try:
            path = self._resolve_path(file_path)
            
            if not path.exists():
                return f"Error: File not found: {file_path}"
            
            if not file_path.endswith('.py'):
                return f"Error: Only Python files are supported for outlining. File: {file_path}"
            
            with open(path, 'r', encoding='utf-8') as f:
                code = f.read()
            
            try:
                tree = ast.parse(code)
            except SyntaxError as e:
                return f"Error: Syntax error in {file_path}: {str(e)}"
            
            outline = [f"# Code Outline: {Path(file_path).name}", ""]
            
            # Process imports
            imports = []
            for node in ast.walk(tree):
                if isinstance(node, ast.Import):
                    for name in node.names:
                        if name.asname:
                            imports.append(f"import {name.name} as {name.asname}")
                        else:
                            imports.append(f"import {name.name}")
                elif isinstance(node, ast.ImportFrom):
                    module = node.module or ''
                    for name in node.names:
                        if name.asname:
                            imports.append(f"from {module} import {name.name} as {name.asname}")
                        else:
                            imports.append(f"from {module} import {name.name}")
            
            if imports:
                outline.append("## Imports")
                for imp in sorted(set(imports)):
                    outline.append(f"- `{imp}`")
                outline.append("")
            
            # Process classes
            classes = []
            for node in ast.walk(tree):
                if isinstance(node, ast.ClassDef):
                    class_info = {
                        'name': node.name,
                        'line': node.lineno,
                        'docstring': ast.get_docstring(node) or '',
                        'methods': []
                    }
                    
                    # Get base classes
                    if node.bases:
                        bases = []
                        for base in node.bases:
                            if isinstance(base, ast.Name):
                                bases.append(base.id)
                            else:
                                bases.append("...")
                        class_info['bases'] = bases
                    
                    # Get methods
                    for child in node.body:
                        if isinstance(child, ast.FunctionDef):
                            method_info = {
                                'name': child.name,
                                'line': child.lineno,
                                'docstring': ast.get_docstring(child) or '',
                                'args': [arg.arg for arg in child.args.args]
                            }
                            class_info['methods'].append(method_info)
                    
                    classes.append(class_info)
            
            if classes:
                outline.append("## Classes")
                for cls in sorted(classes, key=lambda x: x['line']):
                    if 'bases' in cls and cls['bases']:
                        outline.append(f"### {cls['name']}({', '.join(cls['bases'])}) - Line {cls['line']}")
                    else:
                        outline.append(f"### {cls['name']} - Line {cls['line']}")
                    
                    if cls['docstring']:
                        outline.append(f"> {cls['docstring'].split('.')[0]}.")
                    
                    for method in sorted(cls['methods'], key=lambda x: x['line']):
                        args_str = ', '.join(method['args'])
                        outline.append(f"- **{method['name']}({args_str})** - Line {method['line']}")
                        if method['docstring']:
                            outline.append(f"  > {method['docstring'].split('.')[0]}.")
                
                outline.append("")
            
            # Process standalone functions
            functions = []
            for node in ast.iter_child_nodes(tree):
                if isinstance(node, ast.FunctionDef):
                    func_info = {
                        'name': node.name,
                        'line': node.lineno,
                        'docstring': ast.get_docstring(node) or '',
                        'args': [arg.arg for arg in node.args.args]
                    }
                    functions.append(func_info)
            
            if functions:
                outline.append("## Functions")
                for func in sorted(functions, key=lambda x: x['line']):
                    args_str = ', '.join(func['args'])
                    outline.append(f"### {func['name']}({args_str}) - Line {func['line']}")
                    if func['docstring']:
                        outline.append(f"> {func['docstring'].split('.')[0]}.")
                outline.append("")
            
            # Process global variables
            variables = []
            for node in ast.iter_child_nodes(tree):
                if isinstance(node, ast.Assign) and all(isinstance(target, ast.Name) for target in node.targets):
                    for target in node.targets:
                        if isinstance(target, ast.Name):
                            var_info = {
                                'name': target.id,
                                'line': node.lineno
                            }
                            variables.append(var_info)
            
            if variables:
                outline.append("## Global Variables")
                for var in sorted(variables, key=lambda x: x['line']):
                    outline.append(f"- **{var['name']}** - Line {var['line']}")
            
            return "\n".join(outline)
            
        except Exception as e:
            error_msg = f"Error generating outline for {file_path}: {str(e)}"
            logger.error(error_msg)
            return error_msg


if __name__ == '__main__':
    # Simple test
    tool = CodeTool()
    
    # Create a sample Python file
    sample_code = '''
import os
import sys
from datetime import datetime

GLOBAL_VAR = "This is a global variable"

def sample_function(arg1, arg2):
    """Sample function that does something."""
    return arg1 + arg2

class SampleClass:
    """Sample class demonstrating structure."""
    def __init__(self, value):
        self.value = value
        
    def get_value(self):
        """Return the stored value."""
        return self.value
    '''
    
    with open("sample_code.py", "w") as f:
        f.write(sample_code)
    
    # Test different functions
    print("Code Outline:")
    print(tool.get_code_outline("sample_code.py"))
    
    print("\nCode Analysis:")
    print(tool.analyze_code("sample_code.py"))
    
    print("\nFind Symbols:")
    print(tool.find_symbols("sample_code.py", symbol_type="function"))
    
    # Clean up
    import os
    if os.path.exists("sample_code.py"):
        os.remove("sample_code.py") 


# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
part of the code from https://github.com/phidatahq/phidata
"""
from dataclasses import dataclass
from typing import Optional, Dict, List, Tuple, Any

from agentica.emb.base import Emb
import numpy as np
import platform

try:
    from sentence_transformers import SentenceTransformer

    if platform.system() == "Windows":
        numpy_version = np.__version__
        if numpy_version.startswith("2"):
            raise RuntimeError(
                "Incompatible NumPy version detected. Please install NumPy 1.x by running 'pip install numpy<2'."
            )
except ImportError:
    raise ImportError("sentence-transformers not installed, please run `pip install sentence-transformers`")


@dataclass
class SentenceTransformerEmb(Emb):
    model: str = "sentence-transformers/all-MiniLM-L6-v2"
    dimensions: Optional[int] = 384
    client: SentenceTransformer = None
    client_params: Optional[Dict[str, Any]] = None

    @property
    def get_client(self) -> SentenceTransformer:
        if self.client:
            return self.client

        _client_params: Dict[str, Any] = {}
        if self.model:
            _client_params["model_name_or_path"] = self.model
        if self.client_params:
            _client_params.update(self.client_params)

        self.client = SentenceTransformer(**_client_params)
        dim = self.client.get_sentence_embedding_dimension()
        if dim:
            self.dimensions = dim
        return self.client

    def get_embedding(self, text: str, normalize_embeddings: bool = True) -> List[float]:
        # Calculate emb of the text
        return self.get_client.encode([text], normalize_embeddings=normalize_embeddings).tolist()[0]

    def get_embeddings(self, texts: List[str], normalize_embeddings: bool = True) -> List[List[float]]:
        # Calculate emb of the texts
        return self.get_client.encode(texts, normalize_embeddings=normalize_embeddings).tolist()



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: String manipulation tools
"""
import json
from agentica.tools.base import Tool
from agentica.utils.log import logger


class StringTool(Tool):
    def __init__(
            self,
            enable_reverse_string=True,
            enable_text_length=True,
    ):
        super().__init__(name="string_tool")
        if enable_reverse_string:
            self.register(self.reverse_string)
        if enable_text_length:
            self.register(self.text_length)

    def reverse_string(self, s: str) -> str:
        """Reverse the input string.

        Args:
            s (str): Input string.

        Returns:
            str: JSON string of the result.
        """
        try:
            reversed_str = s[::-1]
            logger.debug(f"Reversing string: {s}, Reversed string: {reversed_str}")
            return json.dumps({"operation": "reverse_string", "result": reversed_str})
        except Exception as e:
            logger.error(f"Error in reversing string: {str(e)}")
            return f"Error: {str(e)}"

    def text_length(self, s: str) -> str:
        """Calculate the length of the input string.

        Args:
            s (str): Input string.

        Returns:
            str: JSON string of the result.
        """
        try:
            length = len(s)
            logger.debug(f"Calculating text length: {s}, Text length: {length}")
            return json.dumps({"operation": "text_length", "result": str(length)})
        except Exception as e:
            logger.error(f"Error in calculating text length: {str(e)}")
            return f"Error: {str(e)}"


if __name__ == '__main__':
    tool = StringTool()
    print(tool.reverse_string("Hello, World!"))
    print(tool.text_length("Hello, World!"))



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
from dataclasses import dataclass
import csv
from pathlib import Path
from typing import Any, Dict

from agentica.file.base import File
from agentica.utils.log import logger
from agentica.utils.misc import dataclass_to_dict


@dataclass
class CsvFile(File):
    type: str = "CSV"

    def get_metadata(self) -> Dict[str, Any]:
        if self.name is None:
            self.name = Path(self.data_path).name

        if self.columns is None:
            try:
                # Get the columns from the file
                with open(self.data_path) as f:
                    dict_reader = csv.DictReader(f)
                    if dict_reader.fieldnames is not None:
                        self.columns = list(dict_reader.fieldnames)
            except Exception as e:
                logger.warning(f"Error getting columns from file: {e}")

        return dataclass_to_dict(self, exclude_none=True)



# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
from __future__ import annotations
import datetime
import asyncio
import io
import json
import os
import random
import re
import urllib.parse
from copy import deepcopy
from typing import (
    Any,
    BinaryIO,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    Union,
    cast,
)

from PIL import Image, ImageDraw, ImageFont

from agentica.tools.video_analysis_tool import VideoAnalysisTool
from agentica.model.base import Model
from agentica.model.message import UserMessage
from agentica.model.openai.chat import OpenAIChat
from agentica.agent import Agent
from agentica.tools.base import Tool
from agentica.utils.log import logger
from agentica.config import AGENTICA_HOME
from agentica.utils.filename import sanitize_filename

# Constants
TOP_NO_LABEL_ZONE = 20

WEB_AGENT_SYSTEM_PROMPT = """
You are a helpful web agent that can assist users in browsing the web.
Given a high-level task, you can leverage predefined browser tools to help
users achieve their goals.
        """

PLANNING_AGENT_SYSTEM_PROMPT = """
You are a helpful planning agent that can assist users in planning complex
tasks which need multi-step browser interaction.
        """

OBSERVE_PROMPT_TEMPLATE = """
Please act as a web agent to help me complete the following high-level task:
<task>{task_prompt}</task>
Now, I have made screenshot (only the current viewport, not the full webpage)
based on the current browser state, and marked interactive elements in the
webpage.
Please carefully examine the requirements of the task, and current state of
the browser, and provide the next appropriate action to take.

{detailed_plan_prompt}

Here are the current available browser functions you can use:
{AVAILABLE_ACTIONS_PROMPT}

Here are the latest {history_window} trajectory (at most) you have taken:
<history>
{history}
</history>

Your output should be in json format, including the following fields:
- `observation`: The detailed image description about the current viewport. Do
not over-confident about the correctness of the history actions. You should
always check the current viewport to make sure the correctness of the next
action.
- `reasoning`: The reasoning about the next action you want to take, and the
possible obstacles you may encounter, and how to solve them. Do not forget to
check the history actions to avoid the same mistakes.
- `action_code`: The action code you want to take. It is only one step action
code, without any other texts (such as annotation)

Here is two example of the output:
```json
{{
    "observation": [IMAGE_DESCRIPTION],
    "reasoning": [YOUR_REASONING],
    "action_code": "fill_input_id([ID], [TEXT])"
}}

{{
    "observation":  "The current page is a CAPTCHA verification page on Amazon. It asks the user to ..",
    "reasoning": "To proceed with the task of searching for products, I need to complete..",
    "action_code": "fill_input_id(3, 'AUXPMR')"
}}

Here are some tips for you:
- Never forget the overall question: **{task_prompt}**
- Maybe after a certain operation (e.g. click_id), the page content has not
changed. You can check whether the action step is successful by looking at the
`success` of the action step in the history. If successful, it means that the
page content is indeed the same after the click. You need to try other methods.
- If using one way to solve the problem is not successful, try other ways.
Make sure your provided ID is correct!
- Some cases are very complex and need to be achieve by an iterative process.
You can use the `back()` function to go back to the previous page to try other
methods.
- There are many links on the page, which may be useful for solving the
problem. You can use the `click_id()` function to click on the link to see if
it is useful.
- Always keep in mind that your action must be based on the ID shown in the
current image or viewport, not the ID shown in the history.
- Do not use `stop()` lightly. Always remind yourself that the image only
shows a part of the full page. If you cannot find the answer, try to use
functions like `scroll_up()` and `scroll_down()` to check the full content of
the webpage before doing anything else, because the answer or next key step
may be hidden in the content below.
- If the webpage needs human verification, you must avoid processing it.
Please use `back()` to go back to the previous page, and try other ways.
- If you have tried everything and still cannot resolve the issue, please stop
the simulation, and report issues you have encountered.
- Check the history actions carefully, detect whether you have repeatedly made
the same actions or not.
- When dealing with wikipedia revision history related tasks, you need to
think about the solution flexibly. First, adjust the browsing history
displayed on a single page to the maximum, and then make use of the
find_text_on_page function. This is extremely useful which can quickly locate
the text you want to find and skip massive amount of useless information.
- Flexibly use interactive elements like slide down selection bar to filter
out the information you need. Sometimes they are extremely useful.
```
"""  # noqa: E501

GET_FINAL_ANSWER_PROMPT_TEMPLATE = """
We are solving a complex web task which needs multi-step browser interaction. After the multi-step observation, reasoning and acting with web browser, we think that the task is currently solved.
Here are all trajectory we have taken:
<history>{history}</history>
Please find the final answer, or give valuable insights and founds (e.g. if previous actions contain downloading files, your output should include the path of the downloaded file) about the overall task: <task>{task_prompt}</task>
        """  # noqa: E501

TASK_PLANNING_PROMPT_TEMPLATE = """
<task>{task_prompt}</task>
According to the problem above, if we use browser interaction, what is the general process of the interaction after visiting the webpage `{start_url}`? 

Please note that it can be viewed as Partially Observable MDP. Do not over-confident about your plan.
Please first restate the task in detail, and then provide a detailed plan to solve the task.
"""  # noqa: E501

TASK_REPLANNING_PROMPT_TEMPLATE = """
We are using browser interaction to solve a complex task which needs multi-step actions.
Here are the overall task:
<overall_task>{task_prompt}</overall_task>

In order to solve the task, we made a detailed plan previously. Here is the detailed plan:
<detailed plan>{detailed_plan}</detailed plan>

According to the task above, we have made a series of observations, reasonings, and actions. Here are the latest {history_window} trajectory (at most) we have taken:
<history>{history}</history>

However, the task is not completed yet. As the task is partially observable, we may need to replan the task based on the current state of the browser if necessary.
Now please carefully examine the current task planning schema, and our history actions, and then judge whether the task needs to be fundamentally replanned. If so, please provide a detailed replanned schema (including the restated overall task).

Your output should be in json format, including the following fields:
- `if_need_replan`: bool, A boolean value indicating whether the task needs to be fundamentally replanned.
- `replanned_schema`: str, The replanned schema for the task, which should not be changed too much compared with the original one. If the task does not need to be replanned, the value should be an empty string. 
"""  # noqa: E501

AVAILABLE_ACTIONS_PROMPT = """
1. `fill_input_id(identifier: Union[str, int], text: str)`: Fill an input
field (e.g. search box) with the given text and press Enter.
2. `click_id(identifier: Union[str, int])`: Click an element with the given ID.
3. `hover_id(identifier: Union[str, int])`: Hover over an element with the
given ID.
4. `download_file_id(identifier: Union[str, int])`: Download a file with the
given ID. It returns the path to the downloaded file. If the file is
successfully downloaded, you can stop the simulation and report the path to
the downloaded file for further processing.
5. `scroll_to_bottom()`: Scroll to the bottom of the page.
6. `scroll_to_top()`: Scroll to the top of the page.
7. `scroll_up()`: Scroll up the page. It is suitable when you want to see the
elements above the current viewport.
8. `scroll_down()`: Scroll down the page. It is suitable when you want to see
the elements below the current viewport. If the webpage does not change, It
means that the webpage has scrolled to the bottom.
9. `back()`: Navigate back to the previous page. This is useful when you want
to go back to the previous page, as current page is not useful.
10. `stop()`: Stop the action process, because the task is completed or failed
(impossible to find the answer). In this situation, you should provide your
answer in your output.
11. `get_url()`: Get the current URL of the current page.
12. `find_text_on_page(search_text: str)`: Find the next given text on the
current whole page, and scroll the page to the targeted text. It is equivalent
to pressing Ctrl + F and searching for the text, and is powerful when you want
to fast-check whether the current page contains some specific text.
13. `visit_page(url: str)`: Go to the specific url page.
14. `click_blank_area()`: Click a blank area of the page to unfocus the
current element. It is useful when you have clicked an element but it cannot
unfocus itself (e.g. Menu bar) to automatically render the updated webpage.
15. `ask_question_about_video(question: str)`: Ask a question about the
current webpage which contains video, e.g. youtube websites.
"""

ACTION_WITH_FEEDBACK_LIST = [
    'ask_question_about_video',
    'download_file_id',
    'find_text_on_page',
]


# TypedDicts
class DOMRectangle(TypedDict):
    x: Union[int, float]
    y: Union[int, float]
    width: Union[int, float]
    height: Union[int, float]
    top: Union[int, float]
    right: Union[int, float]
    bottom: Union[int, float]
    left: Union[int, float]


class VisualViewport(TypedDict):
    height: Union[int, float]
    width: Union[int, float]
    offsetLeft: Union[int, float]
    offsetTop: Union[int, float]
    pageLeft: Union[int, float]
    pageTop: Union[int, float]
    scale: Union[int, float]
    clientWidth: Union[int, float]
    clientHeight: Union[int, float]
    scrollWidth: Union[int, float]
    scrollHeight: Union[int, float]


class InteractiveRegion(TypedDict):
    tag_name: str
    role: str
    aria_name: str
    v_scrollable: bool
    rects: List[DOMRectangle]


# Helper Functions
def _get_str(d: Any, k: str) -> str:
    r"""Safely retrieve a string value from a dictionary."""
    if k not in d:
        raise KeyError(f"Missing required key: '{k}'")
    val = d[k]
    if isinstance(val, str):
        return val
    raise TypeError(
        f"Expected a string for key '{k}', but got {type(val).__name__}"
    )


def _get_number(d: Any, k: str) -> Union[int, float]:
    r"""Safely retrieve a number (int or float) from a dictionary"""
    val = d[k]
    if isinstance(val, (int, float)):
        return val
    raise TypeError(
        f"Expected a number (int/float) for key "
        f"'{k}', but got {type(val).__name__}"
    )


def _get_bool(d: Any, k: str) -> bool:
    r"""Safely retrieve a boolean value from a dictionary."""
    val = d[k]
    if isinstance(val, bool):
        return val
    raise TypeError(
        f"Expected a boolean for key '{k}', but got {type(val).__name__}"
    )


def _parse_json_output(
        text: str, logger: Any
) -> Dict[str, Any]:  # Added logger argument
    r"""Extract JSON output from a string."""

    markdown_pattern = r'```(?:json)?\\s*(.*?)\\s*```'
    markdown_match = re.search(markdown_pattern, text, re.DOTALL)
    if markdown_match:
        text = markdown_match.group(1).strip()

    triple_quotes_pattern = r'"""(?:json)?\\s*(.*?)\\s*"""'
    triple_quotes_match = re.search(triple_quotes_pattern, text, re.DOTALL)
    if triple_quotes_match:
        text = triple_quotes_match.group(1).strip()

    try:
        return json.loads(text)
    except json.JSONDecodeError:
        try:
            # Attempt to fix common JSON issues like unquoted keys or using
            # single quotes
            # This is a simplified fix, more robust parsing might be needed
            # for complex cases
            fixed_text = re.sub(
                r"(\w+)(?=\s*:)", r'"\1"', text
            )  # Add quotes to keys
            fixed_text = fixed_text.replace("'", '"')  # Replace single quotes
            # with double
            # Handle boolean values not in lowercase
            fixed_text = re.sub(
                r':\s*True', ': true', fixed_text, flags=re.IGNORECASE
            )
            fixed_text = re.sub(
                r':\s*False', ': false', fixed_text, flags=re.IGNORECASE
            )
            # Remove trailing commas
            fixed_text = re.sub(r",\s*([\}\]])", r"\1", fixed_text)

            return json.loads(fixed_text)
        except json.JSONDecodeError:
            # Fallback to regex extraction if strict JSON parsing fails
            result = {}
            try:
                # Extract boolean-like values
                bool_pattern = r'"?(\w+)"?\s*:\s*(true|false)'
                for match in re.finditer(bool_pattern, text, re.IGNORECASE):
                    key, value = match.groups()
                    result[key.strip('"')] = value.lower() == "true"

                # Extract string values
                str_pattern = r'"?(\w+)"?\s*:\s*"([^"]*)"'
                for match in re.finditer(str_pattern, text):
                    key, value = match.groups()
                    result[key.strip('"')] = value

                # Extract numeric values
                num_pattern = r'"?(\w+)"?\s*:\s*(-?\d+(?:\.\d+)?)'
                for match in re.finditer(num_pattern, text):
                    key, value = match.groups()
                    try:
                        result[key.strip('"')] = int(value)
                    except ValueError:
                        result[key.strip('"')] = float(value)

                # Extract empty string values
                empty_str_pattern = r'"?(\w+)"?\s*:\s*""'
                for match in re.finditer(empty_str_pattern, text):
                    key = match.group(1)
                    result[key.strip('"')] = ""

                if result:
                    return result

                logger.warning(
                    f"Failed to parse JSON output after multiple attempts: "
                    f"{text}"
                )
                return {}
            except Exception as e:
                logger.warning(
                    f"Error during regex extraction from JSON-like string: {e}"
                )
                return {}


def _reload_image(image: Image.Image) -> Image.Image:
    buffer = io.BytesIO()
    image.save(buffer, format="PNG")
    buffer.seek(0)
    return Image.open(buffer)


def dom_rectangle_from_dict(rect: Dict[str, Any]) -> DOMRectangle:
    r"""Create a DOMRectangle object from a dictionary."""
    return DOMRectangle(
        x=_get_number(rect, "x"),
        y=_get_number(rect, "y"),
        width=_get_number(rect, "width"),
        height=_get_number(rect, "height"),
        top=_get_number(rect, "top"),
        right=_get_number(rect, "right"),
        bottom=_get_number(rect, "bottom"),
        left=_get_number(rect, "left"),
    )


def interactive_region_from_dict(region: Dict[str, Any]) -> InteractiveRegion:
    r"""Create an :class:`InteractiveRegion` object from a dictionary."""
    typed_rects: List[DOMRectangle] = []
    for rect_data in region[
        "rects"
    ]:  # Renamed rect to rect_data to avoid conflict
        typed_rects.append(dom_rectangle_from_dict(rect_data))

    return InteractiveRegion(
        tag_name=_get_str(region, "tag_name"),
        role=_get_str(region, "role"),
        aria_name=_get_str(region, "aria-name"),
        v_scrollable=_get_bool(region, "v-scrollable"),
        rects=typed_rects,
    )


def visual_viewport_from_dict(viewport: Dict[str, Any]) -> VisualViewport:
    r"""Create a :class:`VisualViewport` object from a dictionary."""
    return VisualViewport(
        height=_get_number(viewport, "height"),
        width=_get_number(viewport, "width"),
        offsetLeft=_get_number(viewport, "offsetLeft"),
        offsetTop=_get_number(viewport, "offsetTop"),
        pageLeft=_get_number(viewport, "pageLeft"),
        pageTop=_get_number(viewport, "pageTop"),
        scale=_get_number(viewport, "scale"),
        clientWidth=_get_number(viewport, "clientWidth"),
        clientHeight=_get_number(viewport, "clientHeight"),
        scrollWidth=_get_number(viewport, "scrollWidth"),
        scrollHeight=_get_number(viewport, "scrollHeight"),
    )


def add_set_of_mark(
        screenshot: Union[bytes, Image.Image, io.BufferedIOBase],
        ROIs: Dict[str, InteractiveRegion],
) -> Tuple[Image.Image, List[str], List[str], List[str]]:
    if isinstance(screenshot, Image.Image):
        return _add_set_of_mark(screenshot, ROIs)

    if isinstance(screenshot, bytes):
        screenshot = io.BytesIO(screenshot)

    image = Image.open(cast(BinaryIO, screenshot))
    comp, visible_rects, rects_above, rects_below = _add_set_of_mark(
        image, ROIs
    )
    image.close()
    return comp, visible_rects, rects_above, rects_below


def _add_set_of_mark(
        screenshot: Image.Image, ROIs: Dict[str, InteractiveRegion]
) -> Tuple[Image.Image, List[str], List[str], List[str]]:
    r"""Add a set of marks to the screenshot.

    Args:
        screenshot (Image.Image): The screenshot to add marks to.
        ROIs (Dict[str, InteractiveRegion]): The regions to add marks to.

    Returns:
        Tuple[Image.Image, List[str], List[str], List[str]]: A tuple
            containing the screenshot with marked ROIs, ROIs fully within the
            images, ROIs located above the visible area, and ROIs located below
            the visible area.
    """
    visible_rects: List[str] = list()
    rects_above: List[str] = list()  # Scroll up to see
    rects_below: List[str] = list()  # Scroll down to see

    fnt = ImageFont.load_default(14)
    base = screenshot.convert("L").convert("RGBA")
    overlay = Image.new("RGBA", base.size)

    draw = ImageDraw.Draw(overlay)
    for r_key in ROIs:  # Renamed r to r_key
        for rect_item in ROIs[r_key]["rects"]:  # Renamed rect to rect_item
            # Empty rectangles
            if (
                    not rect_item
                    or rect_item["width"] == 0
                    or rect_item["height"] == 0
            ):
                continue

            # TODO: add scroll left and right?
            horizontal_center = (rect_item["right"] + rect_item["left"]) / 2.0
            vertical_center = (rect_item["top"] + rect_item["bottom"]) / 2.0
            is_within_horizon = 0 <= horizontal_center < base.size[0]
            is_above_viewport = vertical_center < 0
            is_below_viewport = vertical_center >= base.size[1]

            if is_within_horizon:
                if is_above_viewport:
                    rects_above.append(r_key)
                elif is_below_viewport:
                    rects_below.append(r_key)
                else:  # Fully visible
                    visible_rects.append(r_key)
                    _draw_roi(draw, int(r_key), fnt, rect_item)

    comp = Image.alpha_composite(base, overlay)
    overlay.close()
    return comp, visible_rects, rects_above, rects_below


def _draw_roi(
        draw: ImageDraw.ImageDraw,
        idx: int,
        font: Union[ImageFont.FreeTypeFont, ImageFont.ImageFont],
        # Made Union explicit
        rect: DOMRectangle,
) -> None:
    r"""Draw a ROI on the image.

    Args:
        draw (ImageDraw.ImageDraw): The draw object.
        idx (int): The index of the ROI.
        font (ImageFont.FreeTypeFont | ImageFont.ImageFont): The font.
        rect (DOMRectangle): The DOM rectangle.
    """
    color = _get_random_color(idx)
    text_color = _get_text_color(color)

    roi = ((rect["left"], rect["top"]), (rect["right"], rect["bottom"]))

    label_location = (rect["right"], rect["top"])
    label_anchor = "rb"

    if label_location[1] <= TOP_NO_LABEL_ZONE:
        label_location = (rect["right"], rect["bottom"])
        label_anchor = "rt"

    draw.rectangle(
        roi, outline=color, fill=(color[0], color[1], color[2], 48), width=2
    )

    bbox = draw.textbbox(
        label_location,
        str(idx),
        font=font,
        anchor=label_anchor,
        align="center",
    )
    bbox = (bbox[0] - 3, bbox[1] - 3, bbox[2] + 3, bbox[3] + 3)
    draw.rectangle(bbox, fill=color)

    draw.text(
        label_location,
        str(idx),
        fill=text_color,
        font=font,
        anchor=label_anchor,
        align="center",
    )


def _get_text_color(
        bg_color: Tuple[int, int, int, int],
) -> Tuple[int, int, int, int]:
    r"""Determine the ideal text color (black or white) for contrast.

    Args:
        bg_color: The background color (R, G, B, A).

    Returns:
        A tuple representing black or white color for text.
    """
    luminance = bg_color[0] * 0.3 + bg_color[1] * 0.59 + bg_color[2] * 0.11
    return (0, 0, 0, 255) if luminance > 120 else (255, 255, 255, 255)


def _get_random_color(identifier: int) -> Tuple[int, int, int, int]:
    r"""Generate a consistent random RGBA color based on the identifier.

    Args:
        identifier: The ID used as a seed to ensure color consistency.

    Returns:
        A tuple representing (R, G, B, A) values.
    """
    rnd = random.Random(int(identifier))
    r_val = rnd.randint(0, 255)  # Renamed r to r_val
    g_val = rnd.randint(125, 255)  # Renamed g to g_val
    b_val = rnd.randint(0, 50)  # Renamed b to b_val
    color = [r_val, g_val, b_val]
    # TODO: check why shuffle is needed?
    rnd.shuffle(color)
    color.append(255)
    return cast(Tuple[int, int, int, int], tuple(color))


class BaseBrowser:
    def __init__(
            self,
            headless=True,
            cache_dir: Optional[str] = None,
            channel: Literal["chrome", "msedge", "chromium"] = "chromium",
            cookie_json_path: Optional[str] = None,
    ):
        self.history: List[Any] = []
        self.headless = headless
        self.channel = channel
        self._ensure_browser_installed()
        self.playwright = None
        self.page_history: List[str] = []
        self.cookie_json_path = cookie_json_path

        self.cache_dir = "tmp/" if cache_dir is None else cache_dir
        os.makedirs(self.cache_dir, exist_ok=True)

        abs_dir_path = os.path.dirname(os.path.abspath(__file__))
        page_script_path = os.path.join(abs_dir_path, "page_script.js")

        try:
            with open(page_script_path, "r", encoding='utf-8') as f:
                self.page_script = f.read()
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Page script file not found at path: {page_script_path}"
            )
        self.browser = None
        self.context = None
        self.page = None
        self.page_url = None
        self.web_agent_model = None

    async def init(self) -> None:
        """Initialize the browser asynchronously."""
        from playwright.async_api import async_playwright

        self.playwright = await async_playwright().start()
        self.browser = await self.playwright.chromium.launch(
            headless=self.headless,
            channel=self.channel
        )

        if self.cookie_json_path and os.path.exists(self.cookie_json_path):
            self.context = await self.browser.new_context(
                accept_downloads=True,
                storage_state=self.cookie_json_path
            )
        else:
            self.context = await self.browser.new_context(accept_downloads=True)

        self.page = await self.context.new_page()

    async def _wait_for_load(self, timeout: int = 20) -> None:
        """Wait for page to load asynchronously."""
        try:
            timeout_ms = timeout * 1000
            await self.page.wait_for_load_state("load", timeout=timeout_ms)
            await asyncio.sleep(2)
        except Exception as e:
            logger.error(f"Error during wait for load: {e}")
            raise

    async def click_blank_area(self) -> None:
        """Click a blank area asynchronously."""
        try:
            await self.page.mouse.click(0, 0)
            await self._wait_for_load()
        except Exception as e:
            logger.error(f"Error during click blank area: {e}")
            raise

    async def visit_page(self, url: str) -> None:
        """Visit a page asynchronously."""
        try:
            await self.page.goto(url)
            await self._wait_for_load()
            self.page_url = url
        except Exception as e:
            logger.error(f"Error during visit page: {e}")
            raise

    async def ask_question_about_video(self, question: str) -> str:
        """Ask a question about video asynchronously."""
        current_url = await self.get_url()

        confirmation_message = (
            f"Do you want to analyze the video on the current "
            f"page({current_url})? This operation may take a long time.(y/n): "
        )
        user_confirmation = input(confirmation_message)

        if user_confirmation.lower() not in ['y', 'yes']:
            return "User cancelled the video analysis."

        model = None
        if hasattr(self, 'web_agent_model') and self.web_agent_model is not None:
            model = self.web_agent_model

        video_analyzer = VideoAnalysisTool(model=model)
        result = await video_analyzer.ask_question_about_video(current_url, question)
        return result

    async def get_screenshot(
            self, save_image: bool = False
    ) -> Tuple[Image.Image, Union[str, None]]:
        """Get screenshot asynchronously."""
        image_data = await self.page.screenshot(timeout=60000)
        image = Image.open(io.BytesIO(image_data))

        file_path = None
        if save_image:
            parsed_url = urllib.parse.urlparse(self.page_url)
            url_name = sanitize_filename(str(parsed_url.path), max_length=241)
            timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            file_path = os.path.join(
                self.cache_dir, f"{url_name}_{timestamp}.png"
            )
            with open(file_path, "wb") as f:
                image.save(f, "PNG")

        return image, file_path

    async def capture_full_page_screenshots(
            self, scroll_ratio: float = 0.8
    ) -> List[str]:
        """Capture full page screenshots asynchronously."""
        screenshots = []
        scroll_height_eval = await self.page.evaluate("document.body.scrollHeight")
        scroll_height = cast(float, scroll_height_eval)

        viewport_height = self.page.viewport_size["height"]
        current_scroll_eval = await self.page.evaluate("window.scrollY")
        current_scroll = cast(float, current_scroll_eval)

        max_height = scroll_height - viewport_height
        scroll_step = int(viewport_height * scroll_ratio)
        last_height = 0.0

        while True:
            logger.debug(
                f"Current scroll: {current_scroll}, max_height: "
                f"{max_height}, step: {scroll_step}"
            )

            _, file_path = await self.get_screenshot(save_image=True)
            if file_path is not None:
                screenshots.append(file_path)

            await self.page.evaluate(f"window.scrollBy(0, {scroll_step})")
            await asyncio.sleep(0.5)

            current_scroll_eval = await self.page.evaluate("window.scrollY")
            current_scroll = cast(float, current_scroll_eval)
            if abs(current_scroll - last_height) < viewport_height * 0.1:
                break

            last_height = current_scroll

        return screenshots

    async def get_visual_viewport(self) -> VisualViewport:
        """Get visual viewport asynchronously."""
        try:
            await self.page.evaluate(self.page_script)
        except Exception as e:
            logger.warning(f"Error evaluating page script: {e}")

        visual_viewport_eval = await self.page.evaluate(
            "MultimodalWebSurfer.getVisualViewport();"
        )
        return visual_viewport_from_dict(cast(Dict[str, Any], visual_viewport_eval))

    async def get_interactive_elements(self) -> Dict[str, InteractiveRegion]:
        """Get interactive elements asynchronously."""
        try:
            await self.page.evaluate(self.page_script)
        except Exception as e:
            logger.warning(f"Error evaluating page script: {e}")

        result = cast(
            Dict[str, Dict[str, Any]],
            await self.page.evaluate("MultimodalWebSurfer.getInteractiveRects();"),
        )

        typed_results: Dict[str, InteractiveRegion] = {}
        for k in result:
            typed_results[k] = interactive_region_from_dict(result[k])

        return typed_results

    async def get_som_screenshot(
            self,
            save_image: bool = False,
    ) -> Tuple[Image.Image, Union[str, None]]:
        """Get screenshot with interactive elements marked asynchronously."""
        await self._wait_for_load()
        screenshot, _ = await self.get_screenshot(save_image=False)
        rects = await self.get_interactive_elements()

        file_path = None
        comp, _, _, _ = _add_set_of_mark(screenshot, rects)

        if save_image:
            parsed_url = urllib.parse.urlparse(self.page_url)
            url_name = sanitize_filename(str(parsed_url.path), max_length=241)
            timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            file_path = os.path.join(
                self.cache_dir, f"{url_name}_{timestamp}.png"
            )
            with open(file_path, "wb") as f:
                comp.save(f, "PNG")

        return comp, file_path

    async def scroll_up(self) -> None:
        """Scroll up asynchronously."""
        try:
            await self.page.keyboard.press("PageUp")
            await self._wait_for_load()
        except Exception as e:
            logger.error(f"Error during scroll up: {e}")
            raise

    async def scroll_down(self) -> None:
        """Scroll down asynchronously."""
        try:
            await self.page.keyboard.press("PageDown")
            await self._wait_for_load()
        except Exception as e:
            logger.error(f"Error during scroll down: {e}")
            raise

    async def get_url(self) -> str:
        """Get current URL asynchronously."""
        return self.page.url

    async def click_id(self, identifier: Union[str, int]) -> None:
        """Click element by ID asynchronously."""
        if isinstance(identifier, int):
            identifier = str(identifier)
        target = self.page.locator(f"[__elementId='{identifier}']")

        try:
            await target.wait_for(timeout=5000)
        except Exception as e:
            logger.debug(f"Error during click operation: {e}")
            raise ValueError("No such element.") from None

        await target.scroll_into_view_if_needed()

        new_page = None
        try:
            async with self.page.expect_event("popup", timeout=1000) as page_info:
                box = await target.bounding_box()
                if box is None:
                    logger.warning(
                        f"Bounding box not found for element '{identifier}'. "
                        f"Cannot click."
                    )
                    return
                await self.page.mouse.click(
                    box["x"] + box["width"] / 2,
                    box["y"] + box["height"] / 2
                )
                new_page = await page_info.value

                if new_page:
                    self.page_history.append(deepcopy(self.page.url))
                    self.page = new_page

        except Exception as e:
            logger.debug(f"Error during click operation: {e}")
            pass

        await self._wait_for_load()

    async def extract_url_content(self) -> str:
        """Extract page content asynchronously."""
        content = await self.page.content()
        return content

    async def download_file_id(self, identifier: Union[str, int]) -> str:
        """Download file by ID asynchronously."""
        if isinstance(identifier, int):
            identifier = str(identifier)
        try:
            target = self.page.locator(f"[__elementId='{identifier}']")
        except Exception as e:
            logger.debug(f"Error during download operation: {e}")
            logger.warning(f"Element with identifier '{identifier}' not found.")
            return f"Element with identifier '{identifier}' not found."

        await target.scroll_into_view_if_needed()

        file_path_val = os.path.join(self.cache_dir)
        await self._wait_for_load()

        try:
            async with self.page.expect_download() as download_info:
                await target.click()
                download = await download_info.value
                file_name = download.suggested_filename

                file_path_val = os.path.join(file_path_val, file_name)
                await download.save_as(file_path_val)

            return f"Downloaded file to path '{file_path_val}'."

        except Exception as e:
            logger.debug(f"Error during download operation: {e}")
            return f"Failed to download file with identifier '{identifier}'."

    async def fill_input_id(self, identifier: Union[str, int], text: str) -> str:
        """Fill input by ID asynchronously."""
        if isinstance(identifier, int):
            identifier = str(identifier)

        try:
            target = self.page.locator(f"[__elementId='{identifier}']")
        except Exception as e:
            logger.debug(f"Error during fill operation: {e}")
            logger.warning(f"Element with identifier '{identifier}' not found.")
            return f"Element with identifier '{identifier}' not found."

        await target.scroll_into_view_if_needed()
        await target.focus()
        try:
            await target.fill(text)
        except Exception as e:
            logger.debug(f"Error during fill operation: {e}")
            await target.press_sequentially(text)

        await target.press("Enter")
        await self._wait_for_load()
        return (
            f"Filled input field '{identifier}' with text '{text}' "
            f"and pressed Enter."
        )

    async def scroll_to_bottom(self) -> str:
        """Scroll to bottom asynchronously."""
        try:
            await self.page.evaluate("window.scrollTo(0, document.body.scrollHeight);")
            await self._wait_for_load()
            return "Scrolled to the bottom of the page."
        except Exception as e:
            logger.error(f"Error during scroll to bottom: {e}")
            raise

    async def scroll_to_top(self) -> str:
        """Scroll to top asynchronously."""
        try:
            await self.page.evaluate("window.scrollTo(0, 0);")
            await self._wait_for_load()
            return "Scrolled to the top of the page."
        except Exception as e:
            logger.error(f"Error during scroll to top: {e}")
            raise

    async def hover_id(self, identifier: Union[str, int]) -> str:
        """Hover over element by ID asynchronously."""
        if isinstance(identifier, int):
            identifier = str(identifier)
        try:
            target = self.page.locator(f"[__elementId='{identifier}']")
        except Exception as e:
            logger.debug(f"Error during hover operation: {e}")
            logger.warning(f"Element with identifier '{identifier}' not found.")
            return f"Element with identifier '{identifier}' not found."

        await target.scroll_into_view_if_needed()
        await target.hover()
        await self._wait_for_load()
        return f"Hovered over element with identifier '{identifier}'."

    async def find_text_on_page(self, search_text: str) -> str:
        """Find text on page asynchronously."""
        script = f"""
        (function() {{ 
            let text = "{search_text}";
            let found = window.find(text);
            if (!found) {{
                let elements = document.querySelectorAll("*:not(script):not(style)"); 
                for (let el of elements) {{
                    if (el.innerText && el.innerText.includes(text)) {{
                        el.scrollIntoView({{behavior: "smooth", block: "center"}});
                        el.style.backgroundColor = "yellow";
                        el.style.border = '2px solid red';
                        return true;
                    }}
                }}
                return false;
            }}
            return true;
        }})();
        """
        found_eval = await self.page.evaluate(script)
        found = cast(bool, found_eval)
        await self._wait_for_load()
        if found:
            return f"Found text '{search_text}' on the page."
        else:
            return f"Text '{search_text}' not found on the page."

    async def back(self):
        """Navigate back asynchronously."""
        page_url_before = self.page.url
        await self.page.go_back()

        page_url_after = self.page.url

        if page_url_after == "about:blank":
            await self.visit_page(page_url_before)

        if page_url_before == page_url_after:
            if len(self.page_history) > 0:
                await self.visit_page(self.page_history.pop())

        await asyncio.sleep(1)
        await self._wait_for_load()

    async def close(self):
        """Close browser asynchronously."""
        if self.browser:
            await self.browser.close()
        if self.playwright:
            await self.playwright.stop()

    async def show_interactive_elements(self):
        """Show interactive elements asynchronously."""
        await self.page.evaluate(self.page_script)
        await self.page.evaluate("""
        () => {
            document.querySelectorAll('a, button, input, select, textarea, [tabindex]:not([tabindex="-1"]), [contenteditable="true"]').forEach(el => {
                el.style.border = '2px solid red';
            });
        }
        """)

    async def get_webpage_content(self) -> str:
        """Get webpage content asynchronously."""
        from html2text import html2text

        await self._wait_for_load()
        html_content = await self.page.content()

        markdown_content = html2text(html_content)
        return markdown_content

    def _ensure_browser_installed(self) -> None:
        """Ensure the browser is installed."""
        import platform
        import subprocess
        import sys
        from pathlib import Path

        path = os.path.join(AGENTICA_HOME, "browser")
        cache_dir = Path(path)
        cache_dir.mkdir(parents=True, exist_ok=True)
        cache_file = cache_dir / f"{self.channel}_installed"

        if cache_file.exists():
            import time
            file_age = time.time() - cache_file.stat().st_mtime
            if file_age < 7 * 24 * 60 * 60:  # 7 days in seconds
                return

        try:
            from playwright.sync_api import sync_playwright
            with sync_playwright() as p:
                browser = p.chromium.launch(channel=self.channel)
                browser.close()
                cache_file.touch()
        except Exception:
            logger.info("Installing Chromium browser...")
            try:
                subprocess.run(
                    [
                        sys.executable,
                        "-m",
                        "playwright",
                        "install",
                        self.channel,
                    ],
                    check=True,
                    capture_output=True,
                )
                if platform.system().lower() == "linux":
                    subprocess.run(
                        [
                            sys.executable,
                            "-m",
                            "playwright",
                            "install-deps",
                            self.channel,
                        ],
                        check=True,
                        capture_output=True,
                    )
                logger.info("Chromium browser installation completed")
                # 安装成功后创建缓存标记
                cache_file.touch()
            except subprocess.CalledProcessError as e:
                raise RuntimeError(f"Failed to install browser: {e.stderr}")


class Browser:
    r"""A class for browsing the web and interacting with web pages.

    This class provides methods for browsing the web and interacting with web
    pages.
    """

    def __init__(
            self,
            headless: bool = False,
            cache_dir: Optional[str] = None,
            channel: Literal["chrome", "msedge", "chromium"] = "chromium",
            history_window: int = 5,
            web_agent_model: Optional[Model] = None,
            planning_agent_model: Optional[Model] = None,
            output_language: str = "en",
            cookie_json_path: Optional[str] = None,
    ):
        r"""Initialize the BrowserToolkit instance.

        Args:
            headless (bool): Whether to run the browser in headless mode.
            cache_dir (Union[str, None]): The directory to store cache files.
            channel (Literal["chrome", "msedge", "chromium"]): The browser
                channel to use. Must be one of "chrome", "msedge", or
                "chromium".
            history_window (int): The window size for storing the history of
                actions.
            web_agent_model (Optional[Model]): The model backend
                for the web agent.
            planning_agent_model (Optional[Model]): The model
                backend for the planning agent.
            output_language (str): The language to use for output.
                (default: :obj:`"en`")
            cookie_json_path (Optional[str]): Path to a JSON file containing
                authentication cookies and browser storage state. If provided
                and the file exists, the browser will load this state to
                maintain
                authenticated sessions without requiring manual login.
                (default: :obj:`None`)
        """
        self.browser = BaseBrowser(
            headless=headless,
            cache_dir=cache_dir,
            channel=channel,
            cookie_json_path=cookie_json_path,
        )
        self.browser.web_agent_model = web_agent_model

        self.history_window = history_window
        self.web_agent_model = web_agent_model
        self.planning_agent_model = planning_agent_model
        self.output_language = output_language

        self.history: List[Dict[str, Any]] = []
        self.web_agent: Agent
        self.planning_agent: Agent
        self.web_agent, self.planning_agent = self._initialize_agent(
            web_agent_model, planning_agent_model
        )

    def _reset(self):
        self.web_agent.reset()
        self.planning_agent.reset()
        self.history = []
        os.makedirs(self.browser.cache_dir, exist_ok=True)

    def _initialize_agent(
            self,
            web_agent_model_backend: Optional[Model],
            planning_agent_model_backend: Optional[Model],
    ) -> Tuple[Agent, Agent]:
        if web_agent_model_backend is None:
            web_agent_model_instance = OpenAIChat()
        else:
            web_agent_model_instance = web_agent_model_backend

        if planning_agent_model_backend is None:
            planning_model = OpenAIChat()
        else:
            planning_model = planning_agent_model_backend

        web_agent = Agent(
            system_message=WEB_AGENT_SYSTEM_PROMPT,
            model=web_agent_model_instance,
            output_language=self.output_language,
        )

        planning_agent = Agent(
            system_message=PLANNING_AGENT_SYSTEM_PROMPT,
            model=planning_model,
            output_language=self.output_language,
        )

        return web_agent, planning_agent

    async def _observe(
            self, task_prompt: str, detailed_plan: Optional[str] = None
    ) -> Tuple[str, str, str]:
        """Let agent observe the current environment and get the next action."""
        detailed_plan_prompt_str = ""

        if detailed_plan is not None:
            detailed_plan_prompt_str = f"""
Here is a plan about how to solve the task step-by-step which you must follow:
<detailed_plan>{detailed_plan}<detailed_plan>
            """

        observe_prompt = OBSERVE_PROMPT_TEMPLATE.format(
            task_prompt=task_prompt,
            detailed_plan_prompt=detailed_plan_prompt_str,
            AVAILABLE_ACTIONS_PROMPT=AVAILABLE_ACTIONS_PROMPT,
            history_window=self.history_window,
            history=self.history[-self.history_window:],
        )

        som_screenshot, _ = await self.browser.get_som_screenshot(save_image=True)
        img = _reload_image(som_screenshot)
        message = UserMessage(content=observe_prompt, images=[img])
        self.web_agent.reset()
        resp = await self.web_agent.arun(message)
        resp_content = resp.content

        resp_dict = _parse_json_output(resp_content, logger)
        observation_result: str = resp_dict.get("observation", "")
        reasoning_result: str = resp_dict.get("reasoning", "")
        action_code: str = resp_dict.get("action_code", "")

        if action_code and "(" in action_code and ")" not in action_code:
            action_match = re.search(
                r'"action_code"\s*:\s*[`"]([^`"]*\([^)]*\))[`"]', resp_content
            )
            if action_match:
                action_code = action_match.group(1)
            else:
                logger.warning(f"Incomplete action_code detected: {action_code}")
                if action_code.startswith("fill_input_id("):
                    parts = action_code.split(",", 1)
                    if len(parts) > 1:
                        id_part = parts[0].replace("fill_input_id(", "").strip()
                        action_code = (
                            f"fill_input_id({id_part}, 'Please fill the text here.')"
                        )

        action_code = action_code.replace("`", "").strip()
        return observation_result, reasoning_result, action_code

    async def _act(self, action_code: str) -> Tuple[bool, str]:
        """Execute the action code asynchronously."""

        def _check_if_with_feedback(action_code: str) -> bool:
            for action_with_feedback in ACTION_WITH_FEEDBACK_LIST:
                if action_with_feedback in action_code:
                    return True
            return False

        def _fix_action_code(action_code: str) -> str:
            match = re.match(r'(\w+)\((.*)\)', action_code)
            if not match:
                return action_code

            func_name, args_str = match.groups()
            args = []
            current_arg = ""
            in_quotes = False
            quote_char = None

            for char in args_str:
                if char in ['"', "'"]:
                    if not in_quotes:
                        in_quotes = True
                        quote_char = char
                        current_arg += char
                    elif char == quote_char:
                        in_quotes = False
                        quote_char = None
                        current_arg += char
                    else:
                        current_arg += char
                elif char == ',' and not in_quotes:
                    args.append(current_arg.strip())
                    current_arg = ""
                else:
                    current_arg += char

            if current_arg:
                args.append(current_arg.strip())

            fixed_args = []
            for arg in args:
                if (
                        (arg.startswith('"') and arg.endswith('"'))
                        or (arg.startswith("'") and arg.endswith("'"))
                        or re.match(r'^-?\d+(\.\d+)?$', arg)
                        or re.match(r'^-?\d+\.?\d*[eE][-+]?\d+$', arg)
                        or re.match(r'^0[xX][0-9a-fA-F]+$', arg)
                ):
                    fixed_args.append(arg)
                else:
                    fixed_args.append(f"'{arg}'")

            return f"{func_name}({', '.join(fixed_args)})"

        action_code = _fix_action_code(action_code)
        prefix = "self.browser."
        code = f"{prefix}{action_code}"

        try:
            if _check_if_with_feedback(action_code):
                result = await eval(code)
                return True, result
            else:
                await exec(code)
                return True, "Action was successful."

        except Exception as e:
            return (
                False,
                f"Error while executing the action {action_code}: {e}. "
                f"If timeout, please recheck whether you have provided the "
                f"correct identifier.",
            )

    async def _get_final_answer(self, task_prompt: str) -> str:
        """Get the final answer based on the task prompt and current browser state."""
        prompt = GET_FINAL_ANSWER_PROMPT_TEMPLATE.format(
            history=self.history, task_prompt=task_prompt
        )
        message = UserMessage(content=prompt)
        self.web_agent.reset()
        resp = await self.web_agent.arun(message)
        return resp.content

    async def _task_planning(self, task_prompt: str, start_url: str) -> str:
        """Plan the task based on the given task prompt."""
        planning_prompt = TASK_PLANNING_PROMPT_TEMPLATE.format(
            task_prompt=task_prompt, start_url=start_url
        )

        message = UserMessage(content=planning_prompt)
        self.planning_agent.reset()
        resp = await self.planning_agent.arun(message)
        return resp.content

    async def _task_replanning(
            self, task_prompt: str, detailed_plan: str
    ) -> Tuple[bool, str]:
        """Replan the task based on the given task prompt."""
        replanning_prompt = TASK_REPLANNING_PROMPT_TEMPLATE.format(
            task_prompt=task_prompt,
            detailed_plan=detailed_plan,
            history_window=self.history_window,
            history=self.history[-self.history_window:],
        )
        self.planning_agent.reset()
        resp = await self.planning_agent.arun(replanning_prompt)
        resp_dict = _parse_json_output(resp.content, logger)

        if_need_replan_eval = resp_dict.get("if_need_replan", False)
        if_need_replan = cast(bool, if_need_replan_eval)
        replanned_schema: str = resp_dict.get("replanned_schema", "")

        if if_need_replan:
            return True, replanned_schema
        else:
            return False, replanned_schema

    async def browse_url(
            self, task_prompt: str, start_url: str, round_limit: int = 12
    ) -> str:
        """A powerful toolkit which can simulate the browser interaction."""
        self._reset()
        task_completed = False
        detailed_plan = await self._task_planning(task_prompt, start_url)
        logger.debug(f"Detailed plan: {detailed_plan}")

        await self.browser.init()
        await self.browser.visit_page(start_url)

        for i in range(round_limit):
            observation, reasoning, action_code = await self._observe(
                task_prompt, detailed_plan
            )
            logger.debug(f"Observation: {observation}")
            logger.debug(f"Reasoning: {reasoning}")
            logger.debug(f"Action code: {action_code}")

            if "stop" in action_code:
                task_completed = True
                trajectory_info = {
                    "round": i,
                    "observation": observation,
                    "thought": reasoning,
                    "action": action_code,
                    "action_if_success": True,
                    "info": None,
                    "current_url": await self.browser.get_url(),
                }
                self.history.append(trajectory_info)
                break

            else:
                success, info = await self._act(action_code)
                if not success:
                    logger.warning(f"Error while executing the action: {info}")

                trajectory_info = {
                    "round": i,
                    "observation": observation,
                    "thought": reasoning,
                    "action": action_code,
                    "action_if_success": success,
                    "info": info,
                    "current_url": await self.browser.get_url(),
                }
                self.history.append(trajectory_info)

                if_need_replan, replanned_schema = await self._task_replanning(
                    task_prompt, detailed_plan
                )
                if if_need_replan:
                    detailed_plan = replanned_schema
                    logger.debug(f"Replanned schema: {replanned_schema}")

        if not task_completed:
            simulation_result = f"""
                The task is not completed within the round limit. Please 
                check the last round {self.history_window} information to 
                see if there is any useful information:
                <history>{self.history[-self.history_window:]}</history>
            """
        else:
            simulation_result = await self._get_final_answer(task_prompt)

        await self.browser.close()
        return simulation_result


class BrowserTool(Tool):
    r"""A class for browsing the web and interacting with web pages.

    This class provides methods for browsing the web and interacting with web
    pages.
    """

    def __init__(
            self,
            headless: bool = False,
            cache_dir: Optional[str] = None,
            channel: Literal["chrome", "msedge", "chromium"] = "chromium",
            history_window: int = 5,
            web_agent_model: Optional[Model] = None,
            planning_agent_model: Optional[Model] = None,
            output_language: str = "en",
            cookie_json_path: Optional[str] = None,
    ):
        r"""Initialize the BrowserToolkit instance.

        Args:
            headless (bool): Whether to run the browser in headless mode.
            cache_dir (Union[str, None]): The directory to store cache files.
            channel (Literal["chrome", "msedge", "chromium"]): The browser
                channel to use. Must be one of "chrome", "msedge", or
                "chromium".
            history_window (int): The window size for storing the history of
                actions.
            web_agent_model (Optional[Model]): The model backend
                for the web agent.
            planning_agent_model (Optional[Model]): The model
                backend for the planning agent.
            output_language (str): The language to use for output.
                (default: :obj:`en`)
            cookie_json_path (Optional[str]): Path to a JSON file containing
                authentication cookies and browser storage state. If provided
                and the file exists, the browser will load this state to
                maintain
                authenticated sessions without requiring manual login.
                (default: :obj:`None`)
        """
        super().__init__(name="BrowserTool")
        self._browser = None
        self.headless = headless
        self.cache_dir = cache_dir
        self.channel = channel
        self.history_window = history_window
        self.web_agent_model = web_agent_model
        self.planning_agent_model = planning_agent_model
        self.output_language = output_language
        self.cookie_json_path = cookie_json_path
        self._event_loop = None
        self.register(self.browse_webpage)

    async def _init_browser(self):
        """Initialize the browser instance."""
        if self._browser is None:
            self._browser = Browser(
                headless=self.headless,
                cache_dir=self.cache_dir,
                channel=self.channel,
                history_window=self.history_window,
                web_agent_model=self.web_agent_model,
                planning_agent_model=self.planning_agent_model,
                output_language=self.output_language,
                cookie_json_path=self.cookie_json_path,
            )
            await self._browser.browser.init()

    def browse_webpage(self, task: str, url: str, round_limit: int = 8) -> str:
        """
        Browse a webpage and perform actions based on the task.
        Args:
            task (str): The task to perform on the webpage.
            url (str): The starting URL for browsing.
            round_limit (int): The maximum number of rounds to perform
                actions. Default is 8.
        Returns:
            str: The result of the browsing task.
        """

        async def _async_browse():
            try:
                await self._init_browser()
                result = await self._browser.browse_url(
                    task_prompt=task,
                    start_url=url,
                    round_limit=round_limit
                )
                return result
            except Exception as e:
                logger.error(f"Error during browser task: {e}")
                raise
            finally:
                if self._browser and self._browser.browser:
                    await self._browser.browser.close()
                    self._browser = None

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        try:
            return loop.run_until_complete(_async_browse())
        except Exception as e:
            logger.error(f"Error in browse_webpage: {e}")
            raise
        finally:
            pending = asyncio.all_tasks(loop)
            for task in pending:
                task.cancel()
            loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))


if __name__ == '__main__':
    m = BrowserTool(
        web_agent_model=OpenAIChat(),
        planning_agent_model=OpenAIChat(),
        cookie_json_path="tmp/cookie.json",
    )
    result = m.browse_webpage(
        task="访问中国天气网(http://www.weather.com.cn)，查看北京今天的天气情况，包括温度、空气质量等信息",
        url="https://www.weather.com.cn",
    )
    print(f"Task result: {result}")



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
part of the code from https://github.com/phidatahq/phidata
"""
from dataclasses import dataclass
from typing import Optional, Dict, List, Any, Tuple

from agentica.emb.base import Emb
from agentica.utils.log import logger


@dataclass
class OllamaEmb(Emb):
    model: str = "quentinz/bge-base-zh-v1.5"
    dimensions: int = 384
    host: Optional[str] = "http://localhost:11434"
    timeout: Optional[Any] = None
    options: Optional[Any] = None
    client_kwargs: Optional[Dict[str, Any]] = None
    ollama_client: Optional[Any] = None

    @property
    def client(self) -> Any:
        try:
            from ollama import Client as OllamaClient
        except ImportError:
            raise ImportError("`ollama` not installed, please run `pip install ollama`")

        if self.ollama_client:
            return self.ollama_client

        _ollama_params: Dict[str, Any] = {}
        if self.host:
            _ollama_params["host"] = self.host
        if self.timeout:
            _ollama_params["timeout"] = self.timeout
        if self.client_kwargs:
            _ollama_params.update(self.client_kwargs)
        self.ollama_client = OllamaClient(**_ollama_params)
        return self.ollama_client

    def _response(self, text: str) -> Dict[str, Any]:
        kwargs: Dict[str, Any] = {}
        if self.options is not None:
            kwargs["options"] = self.options

        return self.client.embeddings(prompt=text, model=self.model, **kwargs)  # type: ignore

    def get_embedding(self, text: str) -> List[float]:
        try:
            response = self._response(text=text)
            if response is None:
                return []
            return response.get("embedding", [])
        except Exception as e:
            logger.warning(e)
            return []

    def get_embeddings(self, texts: List[str]) -> List[List[float]]:
        return [self.get_embedding(text) for text in texts]



# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2024 @ CAMEL-AI.org. All Rights Reserved. =========
import base64
import io
import os
import tempfile
from pathlib import Path
from typing import List, Optional, cast
from PIL import Image

from agentica.model.openai.chat import OpenAIChat, OpenAIImageType
from agentica.model.openai.audio import OpenAIAudioModel
from agentica.model.base import Model
from agentica.tools.base import Tool
from agentica.utils.log import logger
from agentica.tools.video_download_tool import (
    VideoDownloaderTool,
    capture_screenshot,
)

VIDEO_QA_PROMPT = """
Analyze the provided video frames and corresponding audio transcription to \
answer the given question(s) thoroughly and accurately.

Instructions:
    1. Visual Analysis:
        - Examine the video frames to identify visible entities.
        - Differentiate objects, species, or features based on key attributes \
such as size, color, shape, texture, or behavior.
        - Note significant groupings, interactions, or contextual patterns \
relevant to the analysis.

    2. Audio Integration:
        - Use the audio transcription to complement or clarify your visual \
observations.
        - Identify names, descriptions, or contextual hints in the \
transcription that help confirm or refine your visual analysis.

    3. Detailed Reasoning and Justification:
        - Provide a brief explanation of how you identified and distinguished \
each species or object.
        - Highlight specific features or contextual clues that informed \
your reasoning.

    4. Comprehensive Answer:
        - Specify the total number of distinct species or object types \
identified in the video.
        - Describe the defining characteristics and any supporting evidence \
from the video and transcription.

    5. Important Considerations:
        - Pay close attention to subtle differences that could distinguish \
similar-looking species or objects 
          (e.g., juveniles vs. adults, closely related species).
        - Provide concise yet complete explanations to ensure clarity.

output language: {output_language} 

**Audio Transcription:**
{audio_transcription}

**Question:**
{question}
"""


class VideoAnalysisTool(Tool):
    r"""A class for analysing videos with vision-language model.

    Args:
        download_directory (Optional[str], optional): The directory where the
            video will be downloaded to. If not provided, video will be stored
            in a temporary directory and will be cleaned up after use.
            (default: :obj:`None`)
        model (Optional[BaseModelBackend], optional): The model to use for
            visual analysis. (default: :obj:`None`)
        use_audio_transcription (bool, optional): Whether to enable audio
            transcription using OpenAI's audio models. Requires a valid OpenAI
            API key. When disabled, video analysis will be based solely on
            visual content. (default: :obj:`False`)
        frame_interval (float, optional): Interval in seconds between frames
            to extract from the video. (default: :obj:`4.0`)
        output_language (str, optional): The language for output responses.
            (default: :obj:`"English"`)
        cookies_path (Optional[str]): The path to the cookies file
            for the video service in Netscape format. (default: :obj:`None`)
        model_name (str, optional): The name of the model to use for video analysis
            image. Defaults to "gpt-4o".
    """

    def __init__(
            self,
            download_directory: Optional[str] = None,
            model: Optional[Model] = None,
            use_audio_transcription: bool = False,
            frame_interval: float = 4.0,
            output_language: str = "en",
            cookies_path: Optional[str] = None,
            model_name: str = "gpt-4o"
    ) -> None:
        super().__init__(name="video_analysis_tool")
        self._cleanup = download_directory is None
        self._temp_files: list[str] = []  # Track temporary files for cleanup
        self._use_audio_transcription = use_audio_transcription
        self.output_language = output_language
        self.frame_interval = frame_interval

        self._download_directory = Path(
            download_directory or tempfile.mkdtemp()
        ).resolve()

        self.video_downloader_toolkit = VideoDownloaderTool(
            download_directory=str(self._download_directory),
            cookies_path=cookies_path,
        )

        try:
            self._download_directory.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            raise ValueError(
                f"Error creating directory {self._download_directory}: {e}"
            )

        logger.debug(f"Video will be downloaded to {self._download_directory}")

        self.vl_model = model
        self.model_name = model_name
        if self.vl_model is None:
            self.vl_model = OpenAIChat(id=self.model_name)

        # Initialize audio models only if audio transcription is enabled
        self.audio_model = None
        if self._use_audio_transcription:
            try:
                self.audio_model = OpenAIAudioModel()
            except Exception as e:
                logger.warning(
                    f"Failed to initialize OpenAIAudioModels: {e}. "
                    "Audio transcription will be disabled."
                )
                self._use_audio_transcription = False

        # Register the tool methods
        self.register(self.ask_question_about_video)

    def __del__(self):
        r"""Clean up temporary directories and files when the object is
        destroyed.
        """
        # Clean up temporary files
        for temp_file in self._temp_files:
            if os.path.exists(temp_file):
                try:
                    os.remove(temp_file)
                    logger.debug(f"Removed temporary file: {temp_file}")
                except OSError as e:
                    logger.warning(
                        f"Failed to remove temporary file {temp_file}: {e}"
                    )

        # Clean up temporary directory if needed
        if self._cleanup and os.path.exists(self._download_directory):
            try:
                import sys

                if getattr(sys, 'modules', None) is not None:
                    import shutil

                    shutil.rmtree(self._download_directory)
                    logger.debug(
                        f"Removed temp directory: {self._download_directory}"
                    )
            except (ImportError, AttributeError):
                # Skip cleanup if interpreter is shutting down
                pass
            except OSError as e:
                logger.warning(
                    f"Failed to remove temporary directory "
                    f"{self._download_directory}: {e}"
                )

    def _extract_audio_from_video(
            self, video_path: str, output_format: str = "mp3"
    ) -> str:
        r"""Extract audio from the video.

        Args:
            video_path (str): The path to the video file.
            output_format (str): The format of the audio file to be saved.
                (default: :obj:`"mp3"`)

        Returns:
            str: The path to the audio file.
        """
        import ffmpeg

        # Handle case where video file doesn't have an extension
        base_path = os.path.splitext(video_path)[0]
        output_path = f"{base_path}.{output_format}"

        try:
            (
                ffmpeg.input(video_path)
                .output(output_path, vn=None, acodec="libmp3lame")
                .run(quiet=True)
            )
            # Track the audio file for cleanup
            self._temp_files.append(output_path)
            return output_path
        except ffmpeg.Error as e:
            error_message = f"FFmpeg-Python failed: {e}"
            logger.error(error_message)
            raise RuntimeError(error_message)

    def _transcribe_audio(self, audio_path: str) -> str:
        r"""Transcribe the audio of the video."""
        # Check if audio transcription is enabled and audio models are
        # available
        if not self._use_audio_transcription or self.audio_model is None:
            logger.warning("Audio transcription is disabled or not available")
            return "No audio transcription available."

        try:
            audio_transcript = self.audio_model.speech_to_text(audio_path)
            if not audio_transcript:
                logger.warning("Audio transcription returned empty result")
                return "No audio transcription available."
            return audio_transcript
        except Exception as e:
            logger.error(f"Audio transcription failed: {e}")
            return "Audio transcription failed."

    def _extract_keyframes(self, video_path: str) -> List[Image.Image]:
        r"""Extract keyframes from a video based on scene changes and
        regular intervals,and return them as PIL.Image.Image objects.

        Args:
            video_path (str): Path to the video file.

        Returns:
            List[Image.Image]: A list of PIL.Image.Image objects representing
                the extracted keyframes.

        Raises:
            ValueError: If no frames could be extracted from the video.
        """
        import cv2
        import numpy as np
        from scenedetect import (  # type: ignore[import-untyped]
            SceneManager,
            open_video,
        )
        from scenedetect.detectors import (  # type: ignore[import-untyped]
            ContentDetector,
        )

        # Get video information
        cap = cv2.VideoCapture(video_path)
        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        fps = cap.get(cv2.CAP_PROP_FPS)
        duration = total_frames / fps if fps > 0 else 0
        cap.release()

        frame_interval = self.frame_interval  # seconds
        # Maximum number of frames to extract to avoid memory issues
        MAX_FRAMES = 100
        # Minimum time difference (in seconds) to consider frames as distinct
        TIME_THRESHOLD = 1.0

        # Calculate the total number of frames to extract
        if duration <= 0 or fps <= 0:
            logger.warning(
                "Invalid video duration or fps, using default frame count"
            )
            num_frames = 10
        else:
            num_frames = max(int(duration / frame_interval), 1)

            if num_frames > MAX_FRAMES:
                frame_interval = duration / MAX_FRAMES
                num_frames = MAX_FRAMES

            logger.info(
                f"Video duration: {duration:.2f}s, target frames: {num_frames}"
                f"at {frame_interval:.2f}s intervals"
            )

        # Use scene detection to extract keyframes
        # Use open_video instead of VideoManager
        video = open_video(video_path)
        scene_manager = SceneManager()
        scene_manager.add_detector(ContentDetector())

        # Detect scenes using the modern API
        scene_manager.detect_scenes(video)

        scenes = scene_manager.get_scene_list()
        keyframes: List[Image.Image] = []

        # If scene detection is successful, prioritize scene change points
        if scenes:
            logger.info(f"Detected {len(scenes)} scene changes")

            if len(scenes) > num_frames:
                scene_indices = np.linspace(
                    0, len(scenes) - 1, num_frames, dtype=int
                )
                selected_scenes = [scenes[i] for i in scene_indices]
            else:
                selected_scenes = scenes

            # Extract frames from scenes
            for scene in selected_scenes:
                try:
                    # Get start time in seconds
                    start_time = scene[0].get_seconds()
                    frame = capture_screenshot(video_path, start_time)
                    keyframes.append(frame)
                except Exception as e:
                    logger.warning(
                        f"Failed to capture frame at scene change"
                        f" {scene[0].get_seconds()}s: {e}"
                    )

        if len(keyframes) < num_frames and duration > 0:
            logger.info(
                f"Scene detection provided {len(keyframes)} frames, "
                f"supplementing with regular interval frames"
            )

            existing_times = []
            if scenes:
                existing_times = [scene[0].get_seconds() for scene in scenes]

            regular_frames = []
            for i in range(num_frames):
                time_sec = i * frame_interval

                is_duplicate = False
                for existing_time in existing_times:
                    if abs(existing_time - time_sec) < TIME_THRESHOLD:
                        is_duplicate = True
                        break

                if not is_duplicate:
                    try:
                        frame = capture_screenshot(video_path, time_sec)
                        regular_frames.append(frame)
                    except Exception as e:
                        logger.warning(
                            f"Failed to capture frame at {time_sec}s: {e}"
                        )

            frames_needed = num_frames - len(keyframes)
            if frames_needed > 0 and regular_frames:
                if len(regular_frames) > frames_needed:
                    indices = np.linspace(
                        0, len(regular_frames) - 1, frames_needed, dtype=int
                    )
                    selected_frames = [regular_frames[i] for i in indices]
                else:
                    selected_frames = regular_frames

                keyframes.extend(selected_frames)

        if not keyframes:
            logger.warning(
                "No frames extracted, falling back to simple interval"
                "extraction"
            )
            for i in range(
                    min(num_frames, 10)
            ):  # Limit to a maximum of 10 frames to avoid infinite loops
                time_sec = i * (duration / 10 if duration > 0 else 6.0)
                try:
                    frame = capture_screenshot(video_path, time_sec)
                    keyframes.append(frame)
                except Exception as e:
                    logger.warning(
                        f"Failed to capture frame at {time_sec}s: {e}"
                    )

        if not keyframes:
            error_msg = (
                f"Failed to extract any keyframes from video: {video_path}"
            )
            logger.error(error_msg)
            raise ValueError(error_msg)

        # Normalize image sizes
        normalized_keyframes = self._normalize_frames(keyframes)

        logger.info(
            f"Extracted and normalized {len(normalized_keyframes)} keyframes"
        )
        return normalized_keyframes

    def _normalize_frames(
            self, frames: List[Image.Image], target_width: int = 512
    ) -> List[Image.Image]:
        r"""Normalize the size of extracted frames.

        Args:
            frames (List[Image.Image]): List of frames to normalize.
            target_width (int): Target width for normalized frames.

        Returns:
            List[Image.Image]: List of normalized frames.
        """
        normalized_frames: List[Image.Image] = []

        for frame in frames:
            # Get original dimensions
            width, height = frame.size

            # Calculate new height, maintaining aspect ratio
            aspect_ratio = width / height
            new_height = int(target_width / aspect_ratio)

            # Resize image
            resized_frame = frame.resize(
                (target_width, new_height), Image.Resampling.LANCZOS
            )

            # Ensure the image has a proper format
            if resized_frame.mode != 'RGB':
                resized_frame = resized_frame.convert('RGB')

            # Create a new image with explicit format
            with io.BytesIO() as buffer:
                resized_frame.save(buffer, format='JPEG')
                buffer.seek(0)
                formatted_frame = Image.open(buffer)
                formatted_frame.load()  # Load the image data

            normalized_frames.append(formatted_frame)

        return normalized_frames

    def ask_question_about_video(
            self,
            video_path: str,
            question: str,
    ) -> str:
        r"""Ask a question about the video.

        Args:
            video_path (str): The path to the video file.
                It can be a local file or a URL (such as Youtube website).
            question (str): The question to ask about the video.

        Returns:
            str: The answer to the question.
        """
        from urllib.parse import urlparse

        parsed_url = urlparse(video_path)
        is_url = all([parsed_url.scheme, parsed_url.netloc])

        downloaded_video_path = None
        try:
            if is_url:
                downloaded_video_path = (
                    self.video_downloader_toolkit.download_video(video_path)
                )
                if not downloaded_video_path or not os.path.exists(
                        downloaded_video_path
                ):
                    raise ValueError(
                        f"Failed to download video from {video_path}"
                    )
                video_path = downloaded_video_path

            if not os.path.exists(video_path):
                raise FileNotFoundError(f"Video file not found: {video_path}")

            audio_transcript = "No audio transcription available."
            if self._use_audio_transcription:
                audio_path = self._extract_audio_from_video(video_path)
                audio_transcript = self._transcribe_audio(audio_path)

            video_frames = self._extract_keyframes(video_path)

            self.vl_model = cast(Model, self.vl_model)
            prompt = VIDEO_QA_PROMPT.format(
                audio_transcription=audio_transcript,
                question=question,
                output_language=self.output_language,
            )

            hybrid_content = []
            hybrid_content.append(
                {
                    "type": "text",
                    "text": prompt,
                }
            )
            if video_frames and len(video_frames) > 0:
                for image in video_frames:
                    if image.format is None:
                        raise ValueError(
                            f"Image's `format` is `None`, please "
                            f"transform the `PIL.Image.Image` to  one of "
                            f"following supported formats, such as "
                            f"{list(OpenAIImageType)}"
                        )

                    image_type: str = image.format.lower()
                    if image_type not in OpenAIImageType:
                        raise ValueError(
                            f"Image type {image.format} "
                            f"is not supported by OpenAI vision model"
                        )
                    with io.BytesIO() as buffer:
                        image.save(fp=buffer, format=image.format)
                        encoded_image = base64.b64encode(buffer.getvalue()).decode(
                            "utf-8"
                        )
                    image_prefix = f"data:image/{image_type};base64,"
                    hybrid_content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"{image_prefix}{encoded_image}"
                            }
                        }
                    )

            messages = [
                {
                    "role": "user",
                    "content": hybrid_content
                }
            ]
            response = self.vl_model.get_client().chat.completions.create(
                model=self.model_name, messages=messages, max_tokens=4000
            )
            if not response or not response.choices:
                logger.error("Model returned empty response")
                return (
                    "Failed to generate an answer. "
                    "The model returned an empty response."
                )

            answer = response.choices[0].message.content
            return answer
        except Exception as e:
            error_message = f"Error processing video: {e}"
            logger.error(error_message)
            return f"Error: {error_message}"

if __name__ == '__main__':
    # demo case
    video_url = (
        "https://sample-videos.com/video321/mp4/720/big_buck_bunny_720p_1mb.mp4"
    )
    video_analysis_tool = VideoAnalysisTool(
        download_directory=None,
        use_audio_transcription=False,
        frame_interval=4.0,
        output_language="zh",
    )
    question = "What is the main subject of this video?"
    answer = video_analysis_tool.ask_question_about_video(
        video_path=video_url,
        question=question,
    )
    print(f"Question: {question}")
    print(f"Answer: {answer}")
    # Clean up temporary files
    del video_analysis_tool
    import gc
    gc.collect()  # Ensure cleanup happens immediately
    logger.info("Video analysis tool demo completed.")



import json
from time import time
from typing import Optional, Any, Dict, List, Union, Sequence
from pydantic import BaseModel, ConfigDict, Field
from agentica.media import AudioResponse
from agentica.utils.log import logger


class MessageReferences(BaseModel):
    """The references added to user message for RAG"""

    # The query used to retrieve the references.
    query: str
    # References (from the vector database or function calls)
    references: Optional[List[Dict[str, Any]]] = None
    # Time taken to retrieve the references.
    time: Optional[float] = None


class Message(BaseModel):
    """Message sent to the Model"""

    # The role of the message author.
    # One of system, user, assistant, or tool.
    role: str
    # The contents of the message. content is required for all messages,
    # and may be null for assistant messages with function calls.
    content: Optional[Union[List[Any], str]] = None
    # An optional name for the participant.
    # Provides the model information to differentiate between participants of the same role.
    name: Optional[str] = None
    # Tool call that this message is responding to.
    tool_call_id: Optional[str] = None
    # The tool calls generated by the model, such as function calls.
    tool_calls: Optional[List[Dict[str, Any]]] = None

    # Additional modalities
    audio: Optional[Any] = None
    images: Optional[Sequence[Any]] = None
    videos: Optional[Sequence[Any]] = None

    # Output from the models
    audio_output: Optional[AudioResponse] = None

    # The thinking content from the model
    thinking: Optional[str] = None
    redacted_thinking: Optional[str] = None

    # Data from the provider we might need on subsequent messages
    provider_data: Optional[Dict[str, Any]] = None

    # --- Data not sent to the Model API ---
    # The reasoning content from the model
    reasoning_content: Optional[str] = None
    # -*- Attributes not sent to the model
    # The name of the tool called
    tool_name: Optional[str] = Field(None, alias="tool_call_name")
    # Arguments passed to the tool
    tool_args: Optional[Any] = Field(None, alias="tool_call_arguments")
    # The error of the tool call
    tool_call_error: Optional[bool] = None
    # If True, the agent will stop executing after this tool call.
    stop_after_tool_call: bool = False

    # Metrics for the message. This is not sent to the Model API.
    metrics: Dict[str, Any] = Field(default_factory=dict)

    # The references added to the message for RAG
    references: Optional[MessageReferences] = None

    # The Unix timestamp the message was created.
    created_at: int = Field(default_factory=lambda: int(time()))

    model_config = ConfigDict(extra="allow", populate_by_name=True, arbitrary_types_allowed=True)

    def get_content_string(self) -> str:
        """Returns the content as a string."""
        if isinstance(self.content, str):
            return self.content
        if isinstance(self.content, list):
            if len(self.content) > 0 and isinstance(self.content[0], dict) and "text" in self.content[0]:
                return self.content[0].get("text", "")
            else:
                return json.dumps(self.content, ensure_ascii=False)
        return ""

    def to_dict(self) -> Dict[str, Any]:
        _dict = self.model_dump(
            exclude_none=True,
            include={"role", "content", "audio", "name", "tool_call_id", "tool_calls"},
        )
        # Manually add the content field even if it is None
        if self.content is None:
            _dict["content"] = None

        return _dict

    def log(self, level: Optional[str] = None):
        """Log the message to the console

        @param level: The level to log the message at. One of debug, info, warning, or error.
            Defaults to debug.
        """
        level = level or "debug"
        level = level.lower()
        if level == "debug":
            _logger = logger.debug
        elif level == "info":
            _logger = logger.info
        elif level == "warning":
            _logger = logger.warning
        elif level == "error":
            _logger = logger.error
        else:
            _logger = logger.debug

        _logger(f"============== {self.role} ==============")
        if self.name:
            _logger(f"Name: {self.name}")
        if self.tool_call_id:
            _logger(f"Tool call Id: {self.tool_call_id}")
        if self.content:
            if isinstance(self.content, str) or isinstance(self.content, list):
                _logger(self.content)
            elif isinstance(self.content, dict):
                _logger(json.dumps(self.content, ensure_ascii=False))
        if self.tool_calls:
            _logger(f"Tool Calls: {json.dumps(self.tool_calls, ensure_ascii=False)}")
        if self.images:
            _logger(f"Images added: {len(self.images)}")
        if self.videos:
            _logger(f"Videos added: {len(self.videos)}")
        if self.audio:
            if isinstance(self.audio, dict):
                _logger(f"Audio files added: {len(self.audio)}")
                if "id" in self.audio:
                    _logger(f"Audio ID: {self.audio['id']}")
                elif "data" in self.audio:
                    _logger("Message contains raw audio data")
            else:
                _logger(f"Audio file added: {self.audio}")

    def content_is_valid(self) -> bool:
        """Check if the message content is valid."""

        return self.content is not None and len(self.content) > 0


class SystemMessage(Message):
    """System message to the model"""
    role: str = "system"


class UserMessage(Message):
    """User message to the model"""
    role: str = "user"


class AssistantMessage(Message):
    """Assistant message from the model"""
    role: str = "assistant"


class ToolMessage(Message):
    """Tool/Function message"""
    role: str = "tool"



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: URL Crawler Tool
"""
import hashlib
import os
import re
import json
from urllib.parse import urlparse

import markdownify
import requests
from bs4 import BeautifulSoup

from agentica.tools.base import Tool
from agentica.utils.log import logger


class UrlCrawlerTool(Tool):
    def __init__(
            self,
            base_dir: str = os.path.curdir,
            max_content_length: int = 8000,
    ):
        super().__init__(name="url_crawler_tool")
        self.base_dir = base_dir
        self.max_content_length = max_content_length
        self.register(self.url_crawl)

    @staticmethod
    def _generate_file_name_from_url(url: str, max_length=255) -> str:
        url_bytes = url.encode("utf-8")
        hash = hashlib.blake2b(url_bytes).hexdigest()
        parsed_url = urlparse(url)
        file_name = os.path.basename(url)
        prefix = f"{parsed_url.netloc}_{file_name}"
        end = hash[:min(8, max_length - len(parsed_url.netloc) - len(file_name) - 1)]
        file_name = f"{prefix}_{end}"
        return file_name

    @staticmethod
    def parse_html_to_markdown(html: str, url: str = None) -> str:
        """Parse HTML to markdown."""
        soup = BeautifulSoup(html, "html.parser")
        title = soup.title.string if soup.title else "No Title"
        # Remove javascript, style blocks, and hyperlinks
        for element in soup(["script", "style", "a"]):
            element.extract()
        # Remove other common irrelevant elements, e.g., nav, footer, etc.
        for element in soup.find_all(["nav", "footer", "aside", "form", "figure"]):
            element.extract()

        # Convert to markdown -- Wikipedia gets special attention to get a clean version of the page
        if isinstance(url, str) and "wikipedia.org" in url:
            body_elm = soup.find("div", {"id": "mw-content-text"})
            title_elm = soup.find("span", {"class": "mw-page-title-main"})

            if body_elm:
                # What's the title
                main_title = title
                if title_elm and len(title_elm) > 0:
                    main_title = title_elm.string
                webpage_text = "# " + main_title + "\n\n" + markdownify.MarkdownConverter().convert_soup(body_elm)
            else:
                webpage_text = markdownify.MarkdownConverter().convert_soup(soup)
        else:
            webpage_text = markdownify.MarkdownConverter().convert_soup(soup)

        # Convert newlines
        webpage_text = re.sub(r"\r\n", "\n", webpage_text)
        webpage_text = re.sub(r"\n{2,}", "\n\n", webpage_text).strip()
        webpage_text = "# " + title + "\n\n" + webpage_text
        return webpage_text

    def _trim_content(self, content: str) -> str:
        """Trim to the maximum allowed length."""
        if len(content) > self.max_content_length:
            truncated = content[: self.max_content_length]
            return truncated + "... (content truncated)"
        return content

    def url_crawl(self, url: str) -> str:
        """Crawl a website and return the content of the website as a json string.

        Args:
            url (str): The URL of the website to read.

        Example:
            from agentica.tools.url_crawler_tool import UrlCrawlerTool
            m = UrlCrawlerTool()
            url = "https://www.jpmorgan.com/insights/global-research/economy/china-economy-cn#section-header#0"
            r = m.url_crawl(url)
            print(url, '\n\n', r)

        Returns:
            str: The content of the website as a json string.
        """
        filename = self._generate_file_name_from_url(url)
        save_path = os.path.realpath(os.path.join(self.base_dir, filename))
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        content = ""
        try:
            logger.info(f"Crawling URL: {url}")
            headers = {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36",
            }
            response = requests.get(url, stream=True, headers=headers, timeout=60, verify=False)
            response.raise_for_status()

            content_type = response.headers.get("content-type", "")
            if "text/html" in content_type:
                # Get the content of the response
                html = "".join(chunk for chunk in response.iter_content(chunk_size=8192, decode_unicode=True))
                content = self.parse_html_to_markdown(html, url)

                with open(save_path, "w", encoding="utf-8") as f:
                    f.write(content)
            else:
                with open(save_path, "wb") as f:
                    for chunk in response.iter_content(chunk_size=8192):
                        f.write(chunk)
                        content += chunk.decode("utf-8", errors="ignore")
            logger.debug(f"Successfully crawled: {url}, saved to: {save_path}")
        except Exception as e:
            logger.debug(f"Failed to crawl: {url}: {e}")
        crawler_result = {
            "url": url,
            "content": self._trim_content(content),
            "save_path": save_path,
        }
        result = json.dumps(crawler_result, ensure_ascii=False)
        return result


if __name__ == '__main__':
    m = UrlCrawlerTool()
    url = "https://www.jpmorgan.com/insights/global-research/economy/china-economy-cn#section-header#0"
    r = m.url_crawl(url)
    print(url, '\n\n', r)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
from dataclasses import dataclass
from typing import List, Optional, Any, Dict
from agentica.utils.misc import dataclass_to_dict


@dataclass
class File:
    name: Optional[str] = None
    description: Optional[str] = None
    columns: Optional[List[str]] = None
    data_path: Optional[str] = None
    type: str = "FILE"

    def get_metadata(self) -> Dict[str, Any]:
        return dataclass_to_dict(self, exclude_none=True)


# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
from agentica.version import __version__  # noqa, isort: skip
from agentica.config import (
    AGENTICA_HOME,
    AGENTICA_DOTENV_PATH,
    AGENTICA_LOG_LEVEL,
    AGENTICA_LOG_FILE,
)  # noqa, isort: skip
from agentica.utils.log import set_log_level_to_debug, logger
from agentica.utils.io import write_audio_to_file
# model
from agentica.model.openai.chat import OpenAIChat
from agentica.model.openai.chat import OpenAIChat as OpenAILLM
from agentica.model.openai.like import OpenAILike
from agentica.model.azure.openai_chat import AzureOpenAIChat
from agentica.model.azure.openai_chat import AzureOpenAIChat as AzureOpenAILLM
from agentica.model.moonshot import Moonshot
from agentica.model.moonshot import Moonshot as MoonshotChat
from agentica.model.moonshot import Moonshot as MoonshotLLM
from agentica.model.deepseek.chat import DeepSeek
from agentica.model.deepseek.chat import DeepSeek as DeepSeekChat
from agentica.model.deepseek.chat import DeepSeek as DeepSeekLLM
from agentica.model.doubao.chat import Doubao
from agentica.model.doubao.chat import Doubao as DoubaoChat
from agentica.model.together.together import Together
from agentica.model.together.together import Together as TogetherChat
from agentica.model.together.together import Together as TogetherLLM
from agentica.model.xai.grok import Grok
from agentica.model.xai.grok import Grok as GrokChat
from agentica.model.yi.chat import Yi
from agentica.model.yi.chat import Yi as YiChat
from agentica.model.yi.chat import Yi as YiLLM
from agentica.model.qwen.chat import Qwen
from agentica.model.zhipuai.chat import ZhipuAI
from agentica.model.zhipuai.chat import ZhipuAI as ZhipuAIChat
from agentica.model.zhipuai.chat import ZhipuAI as ZhipuAILLM

# memory
from agentica.model.base import Model
from agentica.model.message import Message, MessageReferences
from agentica.model.content import Media, Video, Audio, Image
from agentica.model.response import ModelResponse, FileType
from agentica.memory import (
    AgentRun,
    SessionSummary,
    Memory,
    MemoryManager,
    MemoryClassifier,
    MemoryRetrieval,
    MemorySummarizer,
    AgentMemory,
    WorkflowRun,
    WorkflowMemory,
)
from agentica.memorydb import (
    MemoryDb,
    CsvMemoryDb,
    InMemoryDb,
    SqliteMemoryDb,
    PgMemoryDb,
    MemoryRow
)
from agentica.template import PromptTemplate
# rag
from agentica.run_response import (
    RunResponse,
    RunEvent,
    RunResponseExtraData,
    pprint_run_response,
)
# knowledge
from agentica.knowledge.base import Knowledge
from agentica.knowledge.llamaindex_knowledge import LlamaIndexKnowledge
from agentica.knowledge.langchain_knowledge import LangChainKnowledge
# document
from agentica.document import Document
# vectordb
from agentica.vectordb.base import SearchType, Distance, VectorDb
from agentica.vectordb.memory_vectordb import MemoryVectorDb
# emb
from agentica.emb.base import Emb
from agentica.emb.openai_emb import OpenAIEmb
from agentica.emb.azure_openai_emb import AzureOpenAIEmb
from agentica.emb.hash_emb import HashEmb
from agentica.emb.ollama_emb import OllamaEmb
from agentica.emb.together_emb import TogetherEmb
from agentica.emb.fireworks_emb import FireworksEmb
from agentica.emb.zhipuai_emb import ZhipuAIEmb

# file
from agentica.file.base import File
from agentica.file.csv import CsvFile
from agentica.file.txt import TextFile

# storage
from agentica.storage.agent.base import AgentStorage
from agentica.storage.agent.postgres import PgAgentStorage
from agentica.storage.agent.sqlite import SqlAgentStorage
from agentica.storage.agent.json_file import JsonFileAgentStorage
from agentica.storage.agent.yaml_file import YamlFileAgentStorage
from agentica.storage.workflow.base import WorkflowStorage
from agentica.storage.workflow.sqlite import SqlWorkflowStorage
from agentica.storage.workflow.postgres import PgWorkflowStorage

# tool
from agentica.tools.base import Tool, ModelTool, Function, FunctionCall
from agentica.tools.search_serper_tool import SearchSerperTool
from agentica.tools.baidu_search_tool import BaiduSearchTool
from agentica.tools.run_python_code_tool import RunPythonCodeTool
from agentica.tools.image_analysis_tool import ImageAnalysisTool
from agentica.tools.calculator_tool import CalculatorTool
from agentica.tools.dalle_tool import DalleTool
from agentica.tools.file_tool import FileTool
from agentica.tools.hackernews_tool import HackerNewsTool
from agentica.tools.jina_tool import JinaTool
from agentica.tools.shell_tool import ShellTool
from agentica.tools.text_analysis_tool import TextAnalysisTool
from agentica.tools.weather_tool import WeatherTool

# agent
from agentica.agent import Agent
from agentica.agent_session import AgentSession
from agentica.python_agent import PythonAgent
from agentica.workflow import Workflow
from agentica.workflow_session import WorkflowSession

# mcp
from agentica.mcp.config import MCPConfig



# -*- encoding: utf-8 -*-
"""
@author: XuMing(xuming624@qq.com)
@description:
Search papers in DBLP API.
For detail usage of the DBLP API
please refer to https://dblp.org/faq/How+can+I+fetch+DBLP+data.html
"""
import json

import requests

from agentica.tools.base import Tool
from agentica.utils.log import logger


class DblpTool(Tool):
    def __init__(self):
        super().__init__(name="Dblp_tool")
        self.register(self.search_dblp_and_return_articles)

    def search_dblp_and_return_articles(
            self,
            question: str,
            num_results: int = 20,
            start: int = 0,
            num_completion: int = 10,
    ) -> str:
        """Search papers in the DBLP database.

        Args:
            question (`str`):
                The search query string.
            num_results (`int`, defaults to `20`):
                The number of search results to return.
            start (`int`, defaults to `0`):
                The index of the first search result to return.
            num_completion (`int`, defaults to `10`):
                The number of completions to generate.
        Example:
            from agentica.tools.dblp_tool import DblpTool
            m = DblpTool()
            search_results = m.search_dblp_and_return_articles(question="Extreme Learning Machine")
            print(search_results)

        Returns:
            str, a JSON of the articles with title, authors, venue, pages, year, type, DOI, and URL.
        """
        articles = []
        logger.info(f"Searching dblp for: {question}")
        url = "https://dblp.org/search/publ/api"
        params = {
            "q": question,
            "format": "json",
            "h": num_results,
            "f": start,
            "c": num_completion,
        }
        try:
            response = requests.get(url, params=params)
            response.raise_for_status()
            search_results = response.json()

            hits = search_results.get("result", {}).get("hits", {}).get("hit", [])
            for hit in hits:
                info = hit.get("info", {})
                title = info.get("title", "No title available")
                venue = info.get("venue", "No venue available")
                pages = info.get("pages", "No page information")
                year = info.get("year", "Year not specified")
                pub_type = info.get("type", "Type not specified")
                doi = info.get("doi", "No DOI available")
                url = info.get("url", "No URL available")
                authors_info = info.get("authors", {}).get("author", [])
                if isinstance(
                        authors_info,
                        dict,
                ):  # Check if there's only one author in a dict format
                    authors_info = [authors_info]
                authors = ", ".join(
                    [author["text"] for author in authors_info if "text" in author],
                )
                data = {
                    "title": title,
                    "venue": venue,
                    "pages": pages,
                    "year": year,
                    "type": pub_type,
                    "doi": doi,
                    "url": url,
                    "authors": authors,
                }
                articles.append(data)
        except Exception as e:
            logger.error(f"Error processing article: {e}")
        return json.dumps(articles, indent=2, ensure_ascii=False)


if __name__ == '__main__':
    m = DblpTool()
    search_results = m.search_dblp_and_return_articles(question="Extreme Learning Machine")
    print(search_results)
